############################## COPYRIGHT ##############################
# All rights reserved. The authorized use of this data/code is limited to informational and educational purposes only, and NOT for operational or commercial purposes. No part of this zip file 

# including all the data generation files, example data output and code files may be reproduced, distributed, or transmitted in any form or by any means, without the prior written permission of Prof. 

# Siqian Shen (siqian@umich.edu). For permission requests, write to Siqian Shen (siqian@umich.edu), with a subject “Attention: Permissions for Using DR-noshow Data.”
########################################################################

# obj_text.py -- to generate the objective comparison table 
# 	INPUT: 
# 	-- R-list of R values 
# 	-- p0-list of expected number of showing-up
# 	-- n0-number of appointments
# 	-- train_name-file name of training data

# 	OUTPUT:
# 	-- obj_test.xlsx-a spread sheet containing objective values from solving ED(2),ED(k), and ED(n+1) models and also their solutions


from gurobipy import *
import math
import numpy
from numpy import *
from numpy.linalg import *
from numpy.linalg import inv
import csv
import scipy.io as sio
from scipy.stats import norm
from itertools import chain
import xlsxwriter

timelimit = 3600.0

def  SLP(s,mu,sigma,n,N,nv,cu,cw,co,T,q):	
	
	try:

		m1 = Model("lp")
		# m1.params.timeLimit = timelimit
		m1.setParam('OutputFlag',0)
		# m1.setParam('Heuristics',0)
		# m1.setParam('Cuts', 0)
		# m1.setParam('Presolve', 0)
		# m1.setParam('Threads',1)
		x = []
		w = []
		u = []
		
		for i in range(n+1):
			x.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name='x_%s'%i))
		for i in range(N):
			u.append([])
			w.append([])
			for j in range(n):
				w[i].append(m1.addVar(vtype=GRB.CONTINUOUS,obj = cw[j]/N,lb=0.0,name='w_%s_%s'%(i,j)))
				u[i].append(m1.addVar(vtype=GRB.CONTINUOUS,obj= cu[j]/N,lb=0.0,name='u_%s_%s'%(i,j)))
			w[i].append(m1.addVar(vtype=GRB.CONTINUOUS,obj = co/N,lb=0.0,name='w_%s_%s'%(i,n)))


		# Integrate new variables
		m1.update()
		m1.modelSense = GRB.MINIMIZE

		m1.addConstr(quicksum(x[i] for i in range(n))==T[0])
		m1.addConstr(x[n]==0)

		for t in range(N):
			m1.addConstr(w[t][n]-u[t][n-1]==q[t][n-1]*s[t][n-1]+w[t][n-1]+quicksum(x[l] for l in range(n-1))-T)
			m1.addConstr(w[t][0]==0)
	
			for i in range(1,n):
				m1.addConstr(w[t][i]-u[t][i-1]==q[t][i-1]*s[t][i-1]+w[t][i-1]-x[i-1])
		
	
		m1.optimize()
		if m1.status == GRB.status.OPTIMAL:
	
			x = []
			for i in range(0,n):
				x.append(m1.getVarByName('x_%s'%(i)).x)
				
			return (x,m1.runtime,m1.objVal,m1.status)
		else:
			return([],0,0,m1.status)
	
	except GurobiError as e:
		print ('Error reported: %s'% e.message)

def  ED2(su0,sl0,mu,sigma,n,nv,cu,cw,co,T):
	nv0 = nv
	ep = 0.0#0.338	
	pi = zeros((n+1,n+1))
	for i in range(n):
		for j in range(i,n):
			pi[i][j] = -cu[j]+sum(cw[i+1:j+1])
		pi[i][n]=sum(cw[i+1:n])+co
	sl = zeros((n+1,1))
	su = zeros((n+1,1))

	sl[0:n] = sl0[0:n]
	su[0:n] = su0[0:n]
	try:

		m1 = Model("lp")
		
		m1.setParam('OutputFlag',0)
		# m1.setParam('Heuristics',0)
		# m1.setParam('Cuts', 0)
		# m1.setParam('Presolve', 0)
		# m1.setParam('LazyConstraints', 1)
		# m1.setParam('Threads',1)
		m1.setParam('DualReductions',0)
		# m1.setParam('OptimalityTol', 1e-2)
		rho = []
		al = []
		ze = []
		x = []
		phi = []
		tauu = []
		taul = []
		sigm = []
		lamb = []
		psil = []
		psiu = []
		gamma = []
		for i in range(n+1):
			al.append(m1.addVar(vtype=GRB.CONTINUOUS,obj = 1.0, lb=-GRB.INFINITY, name='al_%s'%(i)))
		for i in range(n+1):
			tauu.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,obj = su[i][0], name = 'tauu_%s'%(i)))
		for i in range(n+1):
			taul.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,obj = -sl[i][0], name='taul_%s'%(i)))
		for i in range(n):
			ze.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=-GRB.INFINITY, name = 'ze_%s'%(i)))
		for i in range(n+1):			
			x.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0, name = 'x_%s'%(i)))
			
			sigm.append([])
			lamb.append([])
			psil.append([])
			psiu.append([])
			for j in range(n+1):
				sigm[i].append([])
				lamb[i].append([])
				psiu[i].append([])
				psil[i].append([])
				for k in range(n+1):
					if (j>=k or i<j or i>=k):
						lamb[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'lamb_%s_%s_%s'%(i,j,k)))
					elif (i==n-1 and k==n and 0<=j<=n-1):
						lamb[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'lamb_%s_%s_%s'%(i,j,k)))
					else:
						lamb[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'lamb_%s_%s_%s'%(i,j,k)))
					if (j>k or i<j or i>k):
						sigm[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'sigm_%s_%s_%s'%(i,j,k)))
						psiu[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'pisu_%s_%s_%s'%(i,j,k)))
						psil[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'pisl_%s_%s_%s'%(i,j,k)))
					else:
						sigm[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'sigm_%s_%s_%s'%(i,j,k)))
						psiu[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'pisu_%s_%s_%s'%(i,j,k)))
						psil[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb = 0.0,name = 'pisl_%s_%s_%s'%(i,j,k)))

		for i in range(n):
			rho.append(m1.addVar(vtype=GRB.CONTINUOUS, lb=-GRB.INFINITY,obj=mu[i], name='rho_%s'%(i)))
			gamma.append(m1.addVar(vtype = GRB.CONTINUOUS,lb=-GRB.INFINITY,obj = nv0[i],name = 'gamma_%s'%(i)))
		
		for i in range(n):
			phi.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'phi_%s'%(i)))
			

		m1.update()
		m1.modelSense = GRB.MINIMIZE

		m1.addConstr(quicksum(x[i] for i in range(n))==T[0])
		# m1.addConstr(quicksum(x[i] for i in range(n))<=T[0][0]+ep)
		m1.addConstr(x[n]==0.0)
	
		m1.addConstr(phi[n-1]==0.0)
	
		
		for k in range(n+1):
			for j in range(k,n+1):
		
				tmp2 = quicksum(lamb[l][k][j] for l in range(k,j))
				tmp3 = sum(phi[j:min(j+1,n)])
				m1.addConstr(quicksum(al[l] for l in range(k,j+1)) - quicksum(sigm[l][k][j] for l in range(k,j+1)) + tmp2+ tmp3>= - quicksum(pi[l][j]*x[l] for l in range(k,j+1))-ep)
		
		for i in range(n):
			m1.addConstr(ze[i]-ep<= gamma[i])
			m1.addConstr(taul[i]-tauu[i]-ep<=rho[i])
		
		for k in range(n+1):
			for i in range(k,n):
				for j in range(i,n+1):
					iu1 = min(j-1,i) + 1
					iu2 = min(n-1,max(2*i-j, i-1)) + 1
					il1 = max(k,i-1)
					il2 = max(0,min(2*i-k-1,i))
					tmp1 = quicksum(lamb[l][k][j] for l in range(il1,iu1))
					tmp2 = quicksum(phi[l] for l in range(il2,iu2))
					m1.addConstr(sigm[i][k][j]+sl[i][0]*psil[i][k][j]-su[i][0]*psiu[i][k][j]+ze[i]-sl[i][0]*taul[i]+su[i][0]*tauu[i]-tmp1-tmp2>=0.0-ep)
	
		for k in range(n+1):
			for i in range(k,n):
				for j in range(i,n+1):	
					m1.addConstr(-psil[i][k][j] + psiu[i][k][j] + taul[i] - tauu[i]>= pi[i][j]-ep)
	
		m1.optimize()
		print m1.status
		if m1.status == GRB.status.OPTIMAL:
	
			m1.write('bug.lp')
			x = []

			for i in range(0,n):
				x.append(m1.getVarByName('x_%s'%(i)).x)
	
			return (x,m1.runtime,m1.objVal,m1.status)
		else:
			return([],0,0,m1.status)
	
	except GurobiError as e:
		print ('Error reported: %s'% e.message)

def  EDn(su0,sl0,mu,sigma,n,nv,cu,cw,co,T):
	nv0 = nv
	
	pi = zeros((n+1,n+1))
	for i in range(n):
		for j in range(i,n):
			pi[i][j] = -cu[j]+sum(cw[i+1:j+1])
		pi[i][n]=sum(cw[i+1:n])+co
	sl = zeros((n+1,1))
	su = zeros((n+1,1))
	sl[0:n] = sl0
	su[0:n] = su0
	try:

		m1 = Model("lp")
		m1.setParam('OutputFlag',0)
		rho = []
		al = []
		x = []
		gamma = []
		beta = []
		for i in range(n+1):
			al.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=-GRB.INFINITY,obj = 1.0, name='al_%s'%(i)))
			x.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0, name = 'x_%s'%(i)))
			beta.append([])			
			for j in range(n+1):
				beta[i].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=-GRB.INFINITY,name = 'beta_%s_%s'%(i,j)))
		for i in range(n):
			rho.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=-GRB.INFINITY, obj=mu[i], name='rho_%s'%(i)))
			gamma.append(m1.addVar(vtype = GRB.CONTINUOUS,lb=-GRB.INFINITY,obj = nv0[i][0],name = 'gamma_%s'%(i)))
		

		# Integrate new variables
		m1.update()
		m1.modelSense = GRB.MINIMIZE

		m1.addConstr(quicksum(x[i] for i in range(n))==T[0])
		m1.addConstr(x[n]==0)
		
		for k in range(n+1):
			for j in range(k,n+1):
				m1.addConstr(quicksum(al[i] for i in range(k,j+1))>= quicksum(beta[i][j] for i in range(k,j+1)))
		m1.addConstr(beta[n][n]==0)	

		for i in range(n):
			for j in range(i,n+1):
				m1.addConstr(beta[i][j]>=-pi[i][j]*x[i]-sl[i][0]*rho[i])
				m1.addConstr(beta[i][j]>=-pi[i][j]*x[i]-su[i][0]*rho[i])
				m1.addConstr(beta[i][j]>=-pi[i][j]*x[i]-sl[i][0]*rho[i]-gamma[i]+sl[i][0]*pi[i][j])
				m1.addConstr(beta[i][j]>=-pi[i][j]*x[i]-su[i][0]*rho[i]-gamma[i]+su[i][0]*pi[i][j])
		m1.optimize()
		if m1.status == GRB.status.OPTIMAL:
			m1.write('EDn.lp')			
			x = []
			for i in range(0,n):
				x.append(m1.getVarByName('x_%s'%(i)).x)
				
			return (x,m1.runtime,m1.objVal,m1.status)
		else:
			return([],0,0,m1.status)
	
	except GurobiError as e:
		print ('Error reported: %s'% e.message)

def CP_noVC(su0,sl0,mu,sigma,n,nv,cu,cw,co,T,K):
	nv0 = nv
	
	nv1 = floor(nv0)
	nv2 = ceil(nv0)
	pi = zeros((n+1,n+1))
	for i in range(n):
		for j in range(i,n):
			pi[i][j] = -cu[j]+sum(cw[i+1:j+1])
		pi[i][n]=sum(cw[i+1:n])+co
	sl = zeros((n+1,1))
	su = zeros((n+1,1))
	sl[0:n] = sl0
	su[0:n] = su0
	epsi = 0.5
	t_MP = []
	T_SB = []
	try:
 ###################master problem###########################3
		m1 = Model("lp")	
		m1.setParam('OutputFlag',0)
		rho = []
		delta = m1.addVar(vtype=GRB.CONTINUOUS,lb=-GRB.INFINITY,obj=1.0,name='delta')
		x = []
		gamma = []
		
		for i in range(n+1):
	
			x.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0, name = 'x_%s'%(i)))
		for i in range(n):
			rho.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=-GRB.INFINITY, obj=mu[i], name='rho_%s'%(i)))
			gamma.append(m1.addVar(vtype = GRB.CONTINUOUS,lb=-GRB.INFINITY,obj = nv0[i],name = 'gamma_%s'%(i)))
		


		al = []
		ze = []
	
		phi = []
		tauu = []
		taul = []
		sigm = []
		lamb = []
		psil = []
		psiu = []
	
		for i in range(n+1):
			al.append(m1.addVar(vtype=GRB.CONTINUOUS, lb=-GRB.INFINITY, name='al_%s'%(i)))
		for i in range(n+1):
			tauu.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0, name = 'tauu_%s'%(i)))
		for i in range(n+1):
			taul.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0, name='taul_%s'%(i)))
		for i in range(n):
			ze.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=-GRB.INFINITY, name = 'ze_%s'%(i)))
		for i in range(n+1):			
			
			sigm.append([])
			lamb.append([])
			psil.append([])
			psiu.append([])
			for j in range(n+1):
				sigm[i].append([])
				lamb[i].append([])
				psiu[i].append([])
				psil[i].append([])
				for k in range(n+1):
					if (j>=k or i<j or i>=k):
						lamb[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'lamb_%s_%s_%s'%(i,j,k)))
					elif (i==n-1 and k==n and 0<=j<=n-1):
						lamb[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'lamb_%s_%s_%s'%(i,j,k)))
					else:
						lamb[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'lamb_%s_%s_%s'%(i,j,k)))
					if (j>k or i<j or i>k):
						sigm[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'sigm_%s_%s_%s'%(i,j,k)))
						psiu[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'pisu_%s_%s_%s'%(i,j,k)))
						psil[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'pisl_%s_%s_%s'%(i,j,k)))
					else:
						sigm[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'sigm_%s_%s_%s'%(i,j,k)))
						psiu[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'pisu_%s_%s_%s'%(i,j,k)))
						psil[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb = 0.0,name = 'pisl_%s_%s_%s'%(i,j,k)))

		
		for i in range(n):
			phi.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'phi_%s'%(i)))
	
		# Integrate new variables
		m1.update()
		m1.modelSense = GRB.MINIMIZE

		m1.addConstr(quicksum(x[i] for i in range(n))==T[0])
		m1.addConstr(x[n]==0)
		

		for i in range(n):

			m1.addConstr(delta + quicksum(sl0[l][0]*rho[l] for l in range(n))+(su[i][0]-sl[i][0])*rho[i]+quicksum(nv1[l][0]*gamma[l] for l in range(n))>=0)
			m1.addConstr(delta + quicksum(sl0[l][0]*rho[l] for l in range(n))+quicksum(nv1[l][0]*gamma[l] for l in range(n)) + (nv2[i][0]-nv1[i][0])*gamma[i]>=0)
		
		
		m1.addConstr(quicksum(al[i]+su[i][0]*tauu[i]-sl[i][0]*taul[i] for i in range(n+1)) <= delta)
		m1.addConstr(phi[n-1]==0.0)
	
		
		for k in range(n+1):
			for j in range(k,n+1):
		
				tmp2 = quicksum(lamb[l][k][j] for l in range(k,j))
				tmp3 = sum(phi[j:min(j+1,n)])
				m1.addConstr(quicksum(al[l] for l in range(k,j+1)) - quicksum(sigm[l][k][j] for l in range(k,j+1)) + tmp2+ tmp3>= - quicksum(pi[l][j]*x[l] for l in range(k,j+1)))
		
		for i in range(n):
			m1.addConstr(ze[i]<= gamma[i])
			m1.addConstr(taul[i]-tauu[i]<=rho[i])
		
		for k in range(n+1):
			for i in range(k,n):
				for j in range(i,n+1):
					iu1 = min(j-1,i) + 1
					iu2 = min(n-1,max(2*i-j, i-1)) + 1
					il1 = max(k,i-1)
					il2 = max(0,min(2*i-k-1,i))
					tmp1 = quicksum(lamb[l][k][j] for l in range(il1,iu1))
					tmp2 = quicksum(phi[l] for l in range(il2,iu2))
					m1.addConstr(sigm[i][k][j]+sl[i][0]*psil[i][k][j]-su[i][0]*psiu[i][k][j]+ze[i]-sl[i][0]*taul[i]+su[i][0]*tauu[i]-tmp1-tmp2>=0.0)
	
		for k in range(n+1):
			for i in range(k,n):
				for j in range(i,n+1):	
					m1.addConstr(-psil[i][k][j] + psiu[i][k][j] + taul[i] - tauu[i]>= pi[i][j])


		m1.optimize()

		if m1.status == GRB.status.OPTIMAL:
			t1 = m1.runtime
			t_MP.append(t1)
			x_sol = []
			rho_sol = []
			gamma_sol = []
			delta_sol = m1.getVarByName('delta').x
			for i in range(0,n):
				rho_sol.append(m1.getVarByName('rho_%s'%(i)).x)
				gamma_sol.append(m1.getVarByName('gamma_%s'%i).x)
			for i in range(0,n+1):
				x_sol.append(m1.getVarByName('x_%s'%(i)).x)
			
		else:
			print 'infeasible'
		

 ##########subproblem##################################
		m2 = Model("mip")
		m2.setParam('OutputFlag',0)
		o = []
		t = []
		var_p = []
		var_q = []
		s = []
		for i  in range(n+1):
			o.append([])
			t.append([])
			var_p.append([])
			var_q.append(m2.addVar(vtype=GRB.BINARY,name='q_%s'%i))
			s.append(m2.addVar(vtype=GRB.CONTINUOUS,lb=sl[i][0],ub=su[i][0],name='s_%s'%i))
			for j in range(n+1):
				o[i].append([])
				t[i].append(m2.addVar(vtype=GRB.BINARY, name='t_%s_%s'%(i,j)))
				var_p[i].append([])
				for k in range(n+1):
					o[i][j].append(m2.addVar(vtype = GRB.CONTINUOUS,lb=0.0,name='o_%s_%s_%s'%(i,j,k)))
					var_p[i][j].append(m2.addVar(vtype=GRB.CONTINUOUS, lb = 0.0,name='p_%s_%s_%s'%(i,j,k)))
			
	# Integrate new variables
		m2.update()
		m2.modelSense = GRB.MAXIMIZE
		m2.addConstr(var_q[n]==0)
		m2.addConstr(s[n]==0)
		for i in range(n+1):
			m2.addConstr(quicksum(quicksum(t[k][j] for j in range(i,n+1)) for k in range(i+1))==1)
		for k in range(n+1):
			for j in range(k,n+1):
				for i in range(k,j+1):
					m2.addConstr(var_p[i][k][j]-t[k][j]<=0)
					m2.addConstr(var_p[i][k][j]-var_q[i]<=0)
					m2.addConstr(var_p[i][k][j] - var_q[i]-t[k][j]>=-1)

					m2.addConstr(o[i][k][j]-sl[i][0]*var_p[i][k][j]>=0)
					m2.addConstr(o[i][k][j]-su[i][0]*var_p[i][k][j]<=0)
					m2.addConstr(o[i][k][j]-s[i]+sl[i][0]*(1-var_p[i][k][j])<=0)
					m2.addConstr(o[i][k][j]-s[i]+su[i][0]*(1-var_p[i][k][j])>=0)

		
		for i in range(n-K+1):
			m2.addConstr(quicksum(var_q[j] for j in range(i,i+K-1+1))>=1)
 #####model end##############################################
	
		
		X_list = []
		X_list.append(x)
		sub_list = []

		find = 0
		iteration = 0
		while find==0:
			if t1 > timelimit:
				break;
			iteration += 1
			
			obj = -(quicksum(rho_sol[l]*s[l] for l in range(n))+quicksum(gamma_sol[l]*var_q[l] for l in range(n)))
			for k in range(n+1):
				for j in range(k,n+1):
					for i in range(k,j+1):
						obj += pi[i][j]*o[i][k][j]-x_sol[i]*pi[i][j]*t[k][j]
			m2.setObjective(obj)
			m2.update()
			m2.optimize()

			if m2.status == GRB.status.OPTIMAL:
				T_SB.append(m2.runtime)
				t1 += m2.runtime
				sub_list.append(m2.objVal)
				
				t_sol = []
				o_sol = []
				s_sol = []
				var_q_sol = []
				for i in range(n+1):
					t_sol.append([])
					o_sol.append([])
					s_sol.append(m2.getVarByName('s_%s'%i).x)
					var_q_sol.append(m2.getVarByName('q_%s'%i).x)
					for j in range(n+1):
						t_sol[i].append(m2.getVarByName('t_%s_%s'%(i,j)).x)
						o_sol[i].append([])
						for  k in range(n+1):
							o_sol[i][j].append(m2.getVarByName('o_%s_%s_%s'%(i,j,k)).x)
				if delta_sol >= m2.objVal-1e-4:
					find = 1
				else:
					obj = -(quicksum(rho[l]*s_sol[l] for l in range(n))+quicksum(gamma[l]*var_q_sol[l] for l in range(n)))
					
					for k in range(n+1):
						for j in range(k,n+1):
							for i in range(k,j+1):
								obj += pi[i][j]*o_sol[i][k][j]-x[i]*pi[i][j]*t_sol[k][j]
					m1.addConstr(obj<= delta)
					m1.optimize()
					
					if m1.status == GRB.status.OPTIMAL:
						t1 += m1.runtime
						t_MP.append(m1.runtime)
						x_sol = []
						rho_sol = []
						gamma_sol = []
						delta_sol = m1.getVarByName('delta').x
						for i in range(0,n):
							
							rho_sol.append(m1.getVarByName('rho_%s'%(i)).x)
							gamma_sol.append(m1.getVarByName('gamma_%s'%i).x)
						for i in range(0,n+1):
							x_sol.append(m1.getVarByName('x_%s'%(i)).x)
						X_list.append(x_sol)
			
					else:
						print 'infeasible'
		
			else:
				print 'subproblem: infeasible'
		if t1< timelimit:
			return (x_sol,t1,m1.objVal,m1.status,iteration,t_MP,T_SB,m2.NumConstrs)
		else:
			return ([],t1,0,999,iteration,t_MP,T_SB,m2.NumConstrs)
		
	except GurobiError as e:
		print ('Error reported: %s'% e.message)

def CP(su0,sl0,mu,sigma,n,nv,cu,cw,co,T,K):
	nv0 = nv

	nv1 = floor(nv0*float(n))/float(n)
	nv2 = ceil(nv0*float(n))/float(n)
	
	pi = zeros((n+1,n+1))
	for i in range(n):
		for j in range(i,n):
			pi[i][j] = -cu[j]+sum(cw[i+1:j+1])
		pi[i][n]=sum(cw[i+1:n])+co
	sl = zeros((n+1,1))
	su = zeros((n+1,1))
	sl[0:n] = sl0
	su[0:n] = su0
	epsi = 0.5
	t_MP = []
	T_SB = []
	try:
  #####################master problem###########################3
		m1 = Model("lp")
		m1.setParam('OutputFlag',0)
		m1.setParam('DualReductions',0)
		m1.setParam('InfUnbdInfo',1)
		rho = []
		delta = m1.addVar(vtype=GRB.CONTINUOUS,lb=-GRB.INFINITY,obj=1.0,name='delta')
		x = []
		gamma = []
		
		for i in range(n+1):
	
			x.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0, name = 'x_%s'%(i)))
		for i in range(n):
			rho.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=-GRB.INFINITY, obj=mu[i], name='rho_%s'%(i)))
			gamma.append(m1.addVar(vtype = GRB.CONTINUOUS,lb=-GRB.INFINITY,obj = nv0[i],name = 'gamma_%s'%(i)))

		al = []
		ze = []
	
		phi = []
		tauu = []
		taul = []
		sigm = []
		lamb = []
		psil = []
		psiu = []
	
		for i in range(n+1):
			al.append(m1.addVar(vtype=GRB.CONTINUOUS, lb=-GRB.INFINITY, name='al_%s'%(i)))
		for i in range(n+1):
			tauu.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0, name = 'tauu_%s'%(i)))
		for i in range(n+1):
			taul.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0, name='taul_%s'%(i)))
		for i in range(n):
			ze.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=-GRB.INFINITY, name = 'ze_%s'%(i)))
		for i in range(n+1):			
			
			sigm.append([])
			lamb.append([])
			psil.append([])
			psiu.append([])
			for j in range(n+1):
				sigm[i].append([])
				lamb[i].append([])
				psiu[i].append([])
				psil[i].append([])
				for k in range(n+1):
					if (j>=k or i<j or i>=k):
						lamb[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'lamb_%s_%s_%s'%(i,j,k)))
					elif (i==n-1 and k==n and 0<=j<=n-1):
						lamb[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'lamb_%s_%s_%s'%(i,j,k)))
					else:
						lamb[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'lamb_%s_%s_%s'%(i,j,k)))
					if (j>k or i<j or i>k):
						sigm[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'sigm_%s_%s_%s'%(i,j,k)))
						psiu[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'pisu_%s_%s_%s'%(i,j,k)))
						psil[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'pisl_%s_%s_%s'%(i,j,k)))
					else:
						sigm[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'sigm_%s_%s_%s'%(i,j,k)))
						psiu[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'pisu_%s_%s_%s'%(i,j,k)))
						psil[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb = 0.0,name = 'pisl_%s_%s_%s'%(i,j,k)))

		
		for i in range(n):
			phi.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'phi_%s'%(i)))
		
		
		# Integrate new variables
		m1.update()
		m1.modelSense = GRB.MINIMIZE

		m1.addConstr(quicksum(x[i] for i in range(n))==T[0])
		m1.addConstr(x[n]==0)
		
		
		for i in range(n):

			m1.addConstr(delta + quicksum(sl0[l][0]*rho[l] for l in range(n))+(su[i][0]-sl[i][0])*rho[i]+quicksum(nv1[l][0]*gamma[l] for l in range(n))>=0)
			m1.addConstr(delta + quicksum(sl0[l][0]*rho[l] for l in range(n))+quicksum(nv1[l][0]*gamma[l] for l in range(n)) + (nv2[i][0]-nv1[i][0])*gamma[i]>=0)
	
		m1.addConstr(quicksum(al[i]+su[i][0]*tauu[i]-sl[i][0]*taul[i] for i in range(n+1)) <= delta)
		m1.addConstr(phi[n-1]==0.0)
	
		
		for k in range(n+1):
			for j in range(k,n+1):
		
				tmp2 = quicksum(lamb[l][k][j] for l in range(k,j))
				tmp3 = sum(phi[j:min(j+1,n)])
				m1.addConstr(quicksum(al[l] for l in range(k,j+1)) - quicksum(sigm[l][k][j] for l in range(k,j+1)) + tmp2+ tmp3>= - quicksum(pi[l][j]*x[l] for l in range(k,j+1)))
		
		for i in range(n):
			m1.addConstr(ze[i]<= gamma[i])
			m1.addConstr(taul[i]-tauu[i]<=rho[i])
		
		for k in range(n+1):
			for i in range(k,n):
				for j in range(i,n+1):
					iu1 = min(j-1,i) + 1
					iu2 = min(n-1,max(2*i-j, i-1)) + 1
					il1 = max(k,i-1)
					il2 = max(0,min(2*i-k-1,i))
					tmp1 = quicksum(lamb[l][k][j] for l in range(il1,iu1))
					tmp2 = quicksum(phi[l] for l in range(il2,iu2))
					m1.addConstr(sigm[i][k][j]+sl[i][0]*psil[i][k][j]-su[i][0]*psiu[i][k][j]+ze[i]-sl[i][0]*taul[i]+su[i][0]*tauu[i]-tmp1-tmp2>=0.0)
	
		for k in range(n+1):
			for i in range(k,n):
				for j in range(i,n+1):	
					m1.addConstr(-psil[i][k][j] + psiu[i][k][j] + taul[i] - tauu[i]>= pi[i][j])

		m1.optimize()

		if m1.status == GRB.status.OPTIMAL:
			t1 = m1.runtime
			t_MP.append(t1)
			x_sol = []
			rho_sol = []
			gamma_sol = []
			delta_sol = m1.getVarByName('delta').x
			for i in range(0,n):
				# x_sol.append(m1.getVarByName('x_%s'%(i)).x)
				rho_sol.append(m1.getVarByName('rho_%s'%(i)).x)
				gamma_sol.append(m1.getVarByName('gamma_%s'%i).x)
			for i in range(0,n+1):
				x_sol.append(m1.getVarByName('x_%s'%(i)).x)
			
		else:
			print "initial master problem:"
			print m1.status
			
		

  ############subproblem##################################
		m2 = Model("mip")
		m2.setParam('OutputFlag',0)
		o = []
		t = []
		var_p = []
		var_q = []
		s = []
		for i  in range(n+1):
			o.append([])
			t.append([])
			var_p.append([])
			var_q.append(m2.addVar(vtype=GRB.BINARY,name='q_%s'%i))
			s.append(m2.addVar(vtype=GRB.CONTINUOUS,lb=sl[i][0],ub=su[i][0],name='s_%s'%i))
			for j in range(n+1):
				o[i].append([])
				t[i].append(m2.addVar(vtype=GRB.BINARY, name='t_%s_%s'%(i,j)))
				var_p[i].append([])
				for k in range(n+1):
					o[i][j].append(m2.addVar(vtype = GRB.CONTINUOUS,lb=0.0,name='o_%s_%s_%s'%(i,j,k)))
					var_p[i][j].append(m2.addVar(vtype=GRB.CONTINUOUS, lb = 0.0,name='p_%s_%s_%s'%(i,j,k)))
			
	# Integrate new variables
		m2.update()
		m2.modelSense = GRB.MAXIMIZE
		m2.addConstr(var_q[n]==0)
		m2.addConstr(s[n]==0)
		for i in range(n+1):
			m2.addConstr(quicksum(quicksum(t[k][j] for j in range(i,n+1)) for k in range(i+1))==1)
		for k in range(n+1):
			for j in range(k,n+1):
				for i in range(k,j+1):
					m2.addConstr(var_p[i][k][j]-t[k][j]<=0)
					m2.addConstr(var_p[i][k][j]-var_q[i]<=0)
					m2.addConstr(var_p[i][k][j] - var_q[i]-t[k][j]>=-1)

					m2.addConstr(o[i][k][j]-sl[i][0]*var_p[i][k][j]>=0)
					m2.addConstr(o[i][k][j]-su[i][0]*var_p[i][k][j]<=0)
					m2.addConstr(o[i][k][j]-s[i]+sl[i][0]*(1-var_p[i][k][j])<=0)
					m2.addConstr(o[i][k][j]-s[i]+su[i][0]*(1-var_p[i][k][j])>=0)

		for i in range(n-K+1):
			m2.addConstr(quicksum(var_q[j] for j in range(i,i+K-1+1))>=1)
  		
  		for i in range(n+1):
  			m2.addConstr(quicksum(quicksum(var_p[i][k][j] for j in range(i,n+1)) for k in range(i+1))==var_q[i])
  			m2.addConstr(s[i]-quicksum(quicksum(o[i][k][j] - sl[i][0]*var_p[i][k][j] for j in range(i,n+1)) for k in range(i+1))>=sl[i][0])
  			m2.addConstr(s[i]-quicksum(quicksum(o[i][k][j] - su[i][0]*var_p[i][k][j] for j in range(i,n+1)) for k in range(i+1))<=su[i][0])
  		
  		for k in range(n+1):
  			for j in range(k,n+1):
  				for i in range(k,j-K+2):
  					m2.addConstr(quicksum(var_p[l][k][j] for l in range(i,i+K))>= t[k][j])
  		for i in range(K-1-1,n):
  			m2.addConstr(quicksum(quicksum(var_p[l][k][i] for l in range(i-K+2,i+1)) for k in range(i-K+2+1))+quicksum(var_p[i+1][i+1][j] for j in range(i-K+2+1,n+1))>=quicksum(t[k][i] for k in range(i-K+2+1)))
  		for i in range(n-K+2):
  			m2.addConstr(quicksum(var_p[i][k][i] for k in range(i+1))+quicksum(quicksum(var_p[l][i+1][j] for j in range(i+K-1,n+1)) for l in range(i+1,i+K))>= quicksum(t[i+1][j] for j in range(i+K-1,n+1)))
  #######model end##############################################
	
		
		X_list = []
		X_list.append(x)
		sub_list = []

		find = 0
		iteration = 0
		while find==0:
			if t1>timelimit:
				break

			iteration += 1
			
			obj = -(quicksum(rho_sol[l]*s[l] for l in range(n))+quicksum(gamma_sol[l]*var_q[l] for l in range(n)))
			for k in range(n+1):
				for j in range(k,n+1):
					for i in range(k,j+1):
						obj += pi[i][j]*o[i][k][j]-x_sol[i]*pi[i][j]*t[k][j]
			m2.setObjective(obj)
			m2.update()
			m2.optimize()

			if m2.status == GRB.status.OPTIMAL:
				T_SB.append(m2.runtime)
				t1 += m2.runtime
				sub_list.append(m2.objVal)
				# m1.write('bug.lp')
				t_sol = []
				o_sol = []
				s_sol = []
				var_q_sol = []
				for i in range(n+1):
					t_sol.append([])
					o_sol.append([])
					s_sol.append(m2.getVarByName('s_%s'%i).x)
					var_q_sol.append(m2.getVarByName('q_%s'%i).x)
					for j in range(n+1):
						t_sol[i].append(m2.getVarByName('t_%s_%s'%(i,j)).x)
						o_sol[i].append([])
						for  k in range(n+1):
							o_sol[i][j].append(m2.getVarByName('o_%s_%s_%s'%(i,j,k)).x)
				# print "%s   $$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$"%iteration
				# for i in range(n+1):
				# 	print var_q_sol[i]

				if delta_sol >= m2.objVal-1e-4:
					find = 1
				else:
					obj = -(quicksum(rho[l]*s_sol[l] for l in range(n))+quicksum(gamma[l]*var_q_sol[l] for l in range(n)))
					
					for k in range(n+1):
						for j in range(k,n+1):
							for i in range(k,j+1):
								obj += pi[i][j]*o_sol[i][k][j]-x[i]*pi[i][j]*t_sol[k][j]
					m1.addConstr(obj<= delta)
					m1.optimize()
				
					if m1.status == GRB.status.OPTIMAL:
						t1 += m1.runtime
						t_MP.append(m1.runtime)
						x_sol = []
						rho_sol = []
						gamma_sol = []
						delta_sol = m1.getVarByName('delta').x
						for i in range(0,n):
							
							rho_sol.append(m1.getVarByName('rho_%s'%(i)).x)
							gamma_sol.append(m1.getVarByName('gamma_%s'%i).x)
						for i in range(0,n+1):
							x_sol.append(m1.getVarByName('x_%s'%(i)).x)
						X_list.append(x_sol)
					else:
						print 'infeasible'
		
			else:
				print 'subproblem: infeasible'
		if t1 < timelimit:	
			
			return (x_sol[0:n],t1,m1.objVal,m1.status,iteration,t_MP,T_SB,m2.NumConstrs)
		else:
			return ([],t1,0,999,iteration,t_MP,T_SB,m2.NumConstrs)
	
		
	except GurobiError as e:
		print ('Error reported: %s'% e.message)

def readfile(train_file):
	mat = sio.loadmat(train_file)
	su = mat['su'].astype(float).tolist()
	sl = mat['sl'].astype(float).tolist()
	
	mu = mat['mu'].astype(float)
	s = mat['s'].astype(float)
	sigma = mat['sigma'].astype(float)
	n = mat['n']
	N = mat['N']
	q = mat['q'].astype(float)
	nv = mat['nv'].astype(float)
	return (su,sl,s,mu,sigma,n,N,q,nv)

def main():
	
# INPUT
	R0 = [0.0,0.25,0.5,0.75,1.0]
	p0 = [0.8,0.6]
	n0 = 10

	cu = 0.5*ones((n0,1))
	cw = 1.0*ones((n0,1))
	co = 10.0
	
	

	train_file  = './train_data/10seed_1465265434'
	workbook = xlsxwriter.Workbook('obj_test.xlsx')
	worksheet = workbook.add_worksheet()
	title1 = ['','','ED2', '', '', '', 'ED(k)', '', '', '', '','ED(n+1)']
	title2 = ['R','n - nv =','', 'k=3', '4', '5', '6', '7', '8', '9', '10','']
	for i in range(len(title1)):
		worksheet.write(0,i+1,title1[i])
		worksheet.write(1,i+1,title2[i])

	title3 = ['R','n-nv=','','x1','x2','x3','x4','x5','x6','x7','x8','x9','x10']
	for i in range(len(title3)):
		worksheet.write(15,i+1,title3[i])
	
	row = 2
	row2 = 16
	
	

		
	for ts in range(len(p0)):
		for tR in range(len(R0)):
			R = R0[tR]		
			col = 1
			col20 = 1
			col2 = col20
			
			worksheet.write(row,col,R)
			worksheet.write(row,col+1,p0[ts]*n0)
			col += 2
			
			worksheet.write(row2,col2,R)
			worksheet.write(row2,col2+1,p0[ts]*n0)
			col2 += 2
			
# INPUT	
			train_name = train_file+'_%sR05.mat'%(n0-int(n0*p0[ts]))
			(su,sl,s,mu,sigma,n,N,q,nv) = readfile(train_name)
			T = sum(mu)+R*sqrt(sum(sigma[i]*sigma[i] for i in range(n)))
			
			
			(x_2,time_2,obj_2,status_2) = ED2(su,sl,mu,sigma,n0,nv,cu,cw,co,T)
			

			worksheet.write(row,col,obj_2)
			col += 1
			worksheet.write(row2,col2,'ED2')
			col2 += 1
			for i in range(n):
				worksheet.write(row2,col2, x_2[i])
				col2 += 1
			row2 += 1
			col2 = col20+2


			for k in range(3,11):

				(x,time,objVal,status,iteration,t_MP,T_SB,nc) = CP(su,sl,mu,sigma,n0,nv,cu,cw,co,T,k)
				
				worksheet.write(row,col,objVal)
				col += 1
				
				worksheet.write(row2,col2,'ED(%s)'%k)
				col2 += 1
				for i in range(n):
					worksheet.write(row2,col2, x[i])
					col2 += 1
				row2 += 1
				col2 = col20+2
			

			(x_n,time_n,obj_n,status_n) = EDn(su,sl,mu,sigma,n0,nv,cu,cw,co,T)
			
			
			worksheet.write(row,col,obj_n)
			col += 1
			row += 1
			
			worksheet.write(row2,col2,'ED(n+1)')
			col2 += 1
			for i in range(n):
				worksheet.write(row2,col2, x_n[i])
				col2 += 1
			row2 += 1
			col2 = col20+2


					
	workbook.close()


		
if __name__ == '__main__':
	main()