############################## COPYRIGHT ##############################
# All rights reserved. The authorized use of this data/code is limited to informational and educational purposes only, and NOT for operational or commercial purposes. No part of this zip file 

# including all the data generation files, example data output and code files may be reproduced, distributed, or transmitted in any form or by any means, without the prior written permission of Prof. 

# Siqian Shen (siqian@umich.edu). For permission requests, write to Siqian Shen (siqian@umich.edu), with a subject “Attention: Permissions for Using DR-noshow Data.”
########################################################################

# simulate.py -- to generate out-of-sample statistics
# 	INPUT: 
# 	-- R-list of R values 
# 	-- p0-list of expected number of showing-up
# 	-- n0-number of appointments
# 	-- mat-file name of in-sample data
# 	-- sol_file-file name of solutions

# 	OUTPUT:
# 	-- inSample_sim.xlsx-a spread sheet containing statistics (mean, std, and quantiles) 				from solving ED(2),ED(n+1), and SLP models


from gurobipy import *
import math
import numpy
from numpy import *
from numpy.linalg import *
from numpy.linalg import inv
import csv
import scipy.io as sio
from scipy.stats import norm
from itertools import chain
import xlsxwriter


# function to read out-of-sample data
def readfile(test_path,test_q,test_s):
	mat_s = sio.loadmat(test_path+test_s)
	s_log = mat_s['s_log'].tolist()
	s_norm = mat_s['s_norm'].tolist()
	test_N = mat_s['test_N']
	mat_q = sio.loadmat(test_path + test_q)	
	q1 = mat_q['q1']
	q2 = mat_q['q2']

	return (test_N,s_log,s_norm,q1,q2)


# function to calculate out-of-sample data
def scenario_test(s,q,x,N,n,workbook,sheet):

	cu = 0.5*ones((n,1))
	cw = 1.0*ones((n,1))
	co = 10.0

	
	u = zeros((n,N))
	w = zeros((n,N))
	OT = zeros((1,N))
	Cost = zeros((1,N))
	for j in  range(1,n):
		for i in range(N):
			w[j][i] = max(0,q[i][j-1]*s[i][j-1]+w[j-1][i]-x[j-1])
			u[j-1][i] = max(0,x[j-1]-q[i][j-1]*s[i][j-1]-w[j-1][i])
	for i in range(N):
		u[n-1][i] = max(0,x[n-1] - q[i][n-1]*s[i][n-1]-w[n-1][i])
		OT[0][i] = u[n-1][i] - x[n-1] +q[i][n-1]*s[i][n-1] + w[n-1][i]
		Cost[0][i] = co*OT[0][i] + sum(cu[l]*u[l][i] for l in range(n)) + sum(cw[l]*w[l][i] for l in range(n)) 

	
	worksheet = workbook.add_worksheet(sheet)
	col = 0
	row = 0
	worksheet.write(row,col,'waiting time')
	worksheet. write(row,col+12, 'overtime')
	worksheet.write(row, col+14, 'Idel time')
	col = 1
	row = 1
	for i in range(N):
		col = 1
		worksheet.write(row,0,i)
		for j in range(n):
			worksheet.write(row,col,w[j][i])
			col += 1
		row += 1
	col = 12
	row = 1
	for i in range(N):
		worksheet.write(row,11,i)
		worksheet.write(row,col,OT[0][i])
		row += 1
	col = 15
	row = 1
	for i in range(N):
		worksheet.write(row,14,i)
		col = 15
		for j in range(n):
			worksheet.write(row,col,u[j][i])
			col += 1
		row += 1
	
	TotalWT = sum(w,axis = 0)/float(n-1)
	TotalIT = sum(u,axis = 0)/float(n)
	TotalOT = sum(OT,axis = 0)
	TotalCost = sum(Cost,axis=0)

	AvgWT = mean(TotalWT)
	AvgIT = mean(TotalIT)
	AvgOT = mean(TotalOT)
	AvgCost = mean(TotalCost)
	
	
	stdWT = std(w, ddof = 1)
	stdIT = std(u, ddof = 1)
	stdOT = std(OT, ddof = 1)
	stdCost = std(Cost, ddof = 1)

	quan = [50,75,85,90,95,99]
	
	
	Wpercentile50 = percentile(w,quan[0])
	Ipercentile50 = percentile(u,quan[0])
	Opercentile50 = percentile(OT,quan[0])
	Cost50 = percentile(Cost,quan[0])

	Wpercentile75 = percentile(w,quan[1])
	Ipercentile75 = percentile(u,quan[1])
	Opercentile75 = percentile(OT,quan[1])
	Cost75 = percentile(Cost,quan[1])
	
	Wpercentile85 = percentile(w,quan[2])
	Ipercentile85 = percentile(u,quan[2])
	Opercentile85 = percentile(OT,quan[2])
	Cost85 = percentile(Cost,quan[2])
	
	Wpercentile90 = percentile(w,quan[3])
	Ipercentile90 = percentile(u,quan[3])
	Opercentile90 = percentile(OT,quan[3])
	Cost90 = percentile(Cost,quan[3])
	
	Wpercentile95 = percentile(w,quan[4])
	Ipercentile95 = percentile(u,quan[4])
	Opercentile95 = percentile(OT,quan[4])
	Cost95 = percentile(Cost,quan[4])
	
	Wpercentile99 = percentile(w,quan[5])
	Ipercentile99 = percentile(u,quan[5])
	Opercentile99 = percentile(OT,quan[5])
	Cost99 = percentile(Cost,quan[5])
	
	
	return (stdWT,stdIT,stdOT,AvgWT,AvgIT,AvgOT,
		Wpercentile95,Wpercentile75,Ipercentile95,Ipercentile75,
		Opercentile95,Opercentile75,Wpercentile99,Ipercentile99,Opercentile99,Wpercentile90,Ipercentile90,Opercentile90,
		Wpercentile50,Ipercentile50,Opercentile50,Wpercentile85,Ipercentile85,Opercentile85,
		stdCost, AvgCost, Cost99, Cost95, Cost85, Cost75,Cost50,Cost90)
def main():

	n0 = 10
	R = [0.0,0.25,0.5,0.75,1.0]
	p0 = [0.8,0.6]	
	
	workbook = xlsxwriter.Workbook('cut-out-of-sample_PN0610_2_correct.xlsx')

	title1 = ['','n-nv=0.2','',' ','',' ','','','n-nv=0.4','','']
	title2 = ['WaitT','OverT','IdelT','Cost',' ','Statistic','Model','WaitT','OverT','IdelT','Cost']
	
	

	for tR in range(len(R)):

		for tt in [1]:
			
			worksheet = workbook.add_worksheet('R%s_%s'%(R[tR],tt))
			for i in range(len(title1)):
				worksheet.write(0,2+i,title1[i])
				worksheet.write(1,2+i,title2[i])
			worksheet.write(1,0,'Statistics')
			worksheet.write(1,1,'Model')
		
			col = -7
			row0 = 2
			for ts in range(len(p0)):
				recordfile = 'c_cut-R%sp%srawdata.xlsx'%(R[tR],p0[ts])
				workbook2 =  xlsxwriter.Workbook(recordfile)

				row = row0
				col += 7
				
				mat = sio.loadmat('./test_data/cutrevisedn10seed_%sR_test.mat'%(n0-int(n0*p0[ts])))
				# mat = sio.loadmat('./train_data/10seed_1465265434_%sR05.mat'%(n0-int(n0*p0[ts])))
				s = mat['s'].astype(float)
				q = mat['q'].astype(float)
				N = int(mat['N'])
				
				sol_file = './solution/n%sp%sR%s_%s05.mat'%(n0,p0[ts],R[tR],tt)
				mat_x = sio.loadmat(sol_file)
				x2 = mat_x['x2']
				xn = mat_x['xn']
				xslp = mat_x['xslp']
		
				recordsheet = 'ED2'
				(stdWT,stdIT,stdOT,AvgWT,AvgIT,AvgOT,Wpercentile95,Wpercentile75,Ipercentile95,Ipercentile75,
					Opercentile95,Opercentile75,Wpercentile99,Ipercentile99,Opercentile99,Wpercentile90,Ipercentile90,Opercentile90,Wpercentile50,Ipercentile50,Opercentile50,
					Wpercentile85,Ipercentile85,Opercentile85,stdCost, AvgCost, Cost99, Cost95, Cost85, Cost75,Cost50,Cost90) = scenario_test(s,q,x2[0],N,n0,workbook2,recordsheet)
				
				
				worksheet.write(row,col+1,'ED2')
				worksheet.write(row,col+2,AvgWT)
				worksheet.write(row,col+3,AvgOT)
				worksheet.write(row,col+4,AvgIT)
				worksheet.write(row,col+5,AvgCost)
				row += 4 
				worksheet.write(row,col+1,'ED2')
				worksheet.write(row,col+2,stdWT)
				worksheet.write(row,col+3,stdOT)
				worksheet.write(row,col+4,stdIT)
				worksheet.write(row,col+5,stdCost)
				row += 4
				worksheet.write(row,col+1,'ED2')
				worksheet.write(row,col+2,Wpercentile50)
				worksheet.write(row,col+3,Opercentile50)
				worksheet.write(row,col+4,Ipercentile50)				
				worksheet.write(row,col+5,Cost50)				
				row += 4
				worksheet.write(row,col+1,'ED2')
				worksheet.write(row,col+2,Wpercentile75)
				worksheet.write(row,col+3,Opercentile75)
				worksheet.write(row,col+4,Ipercentile75)
				worksheet.write(row,col+5,Cost75)
				row += 4
				worksheet.write(row,col+1,'ED2')
				worksheet.write(row,col+2,Wpercentile85)
				worksheet.write(row,col+3,Opercentile85)
				worksheet.write(row,col+4,Ipercentile85)
				worksheet.write(row,col+5,Cost85)
				row += 4
				worksheet.write(row,col+1,'ED2')
				worksheet.write(row,col+2,Wpercentile90)
				worksheet.write(row,col+3,Opercentile90)
				worksheet.write(row,col+4,Ipercentile90)
				worksheet.write(row,col+5,Cost90)
			
				row += 4
				worksheet.write(row,col+1,'ED2')
				worksheet.write(row,col+2,Wpercentile95)
				worksheet.write(row,col+3,Opercentile95)
				worksheet.write(row,col+4,Ipercentile95)
				worksheet.write(row,col+5,Cost95)
				row += 4
				worksheet.write(row,col+1,'ED2')
				worksheet.write(row,col+2,Wpercentile99)
				worksheet.write(row,col+3,Opercentile99)
				worksheet.write(row,col+4,Ipercentile99)
				worksheet.write(row,col+5,Cost99)
				
				row = row0+1
				recordsheet = 'EDn'
				(stdWT,stdIT,stdOT,AvgWT,AvgIT,AvgOT,Wpercentile95,Wpercentile75,Ipercentile95,Ipercentile75,
					Opercentile95,Opercentile75,Wpercentile99,Ipercentile99,Opercentile99,Wpercentile90,Ipercentile90,Opercentile90,Wpercentile50,Ipercentile50,Opercentile50,
					Wpercentile85,Ipercentile85,Opercentile85,stdCost, AvgCost, Cost99, Cost95, Cost85, Cost75,Cost50,Cost90) = scenario_test(s,q,xn[0],N,n0,workbook2,recordsheet)
		
				worksheet.write(row,col+1,'EDn+1')
				worksheet.write(row,col+0,'Mean')
				worksheet.write(row,col+2,AvgWT)
				worksheet.write(row,col+3,AvgOT)
				worksheet.write(row,col+4,AvgIT)
				worksheet.write(row,col+5,AvgCost)
				row += 4
				worksheet.write(row,col+1,'EDn+1')
				worksheet.write(row,col+0,'Std.')
				worksheet.write(row,col+2,stdWT)
				worksheet.write(row,col+3,stdOT)
				worksheet.write(row,col+4,stdIT)
				worksheet.write(row,col+5,stdCost)
				row += 4
				worksheet.write(row,col+1,'EDn+1')
				worksheet.write(row,col+0,'50%')
				worksheet.write(row,col+2,Wpercentile50)
				worksheet.write(row,col+3,Opercentile50)
				worksheet.write(row,col+4,Ipercentile50)
				worksheet.write(row,col+5,Cost50)				
				row += 4
				worksheet.write(row,col+1,'EDn+1')
				worksheet.write(row,col+0,'75%')
				worksheet.write(row,col+2,Wpercentile75)
				worksheet.write(row,col+3,Opercentile75)
				worksheet.write(row,col+4,Ipercentile75)
				worksheet.write(row,col+5,Cost75)				
				row += 4
				worksheet.write(row,col+1,'EDn+1')
				worksheet.write(row,col+0,'85%')
				worksheet.write(row,col+2,Wpercentile85)
				worksheet.write(row,col+3,Opercentile85)
				worksheet.write(row,col+4,Ipercentile85)
				worksheet.write(row,col+5,Cost85)
				row += 4
				worksheet.write(row,col+1,'EDn+1')
				worksheet.write(row,col+0,'90%')
				worksheet.write(row,col+2,Wpercentile90)
				worksheet.write(row,col+3,Opercentile90)
				worksheet.write(row,col+4,Ipercentile90)
				worksheet.write(row,col+5,Cost90)
							
				row += 4
				worksheet.write(row,col+1,'EDn+1')
				worksheet.write(row,col+0,'95%')
				worksheet.write(row,col+2,Wpercentile95)
				worksheet.write(row,col+3,Opercentile95)
				worksheet.write(row,col+4,Ipercentile95)
				worksheet.write(row,col+5,Cost95)				
				row += 4
				worksheet.write(row,col+1,'EDn+1')
				worksheet.write(row,col+0,'99%')
				worksheet.write(row,col+2,Wpercentile99)
				worksheet.write(row,col+3,Opercentile99)
				worksheet.write(row,col+4,Ipercentile99)
				worksheet.write(row,col+5,Cost99)				
				
				row = row0+2
				recordsheet = 'SLP'
				(stdWT,stdIT,stdOT,AvgWT,AvgIT,AvgOT,Wpercentile95,Wpercentile75,Ipercentile95,Ipercentile75,
					Opercentile95,Opercentile75,Wpercentile99,Ipercentile99,Opercentile99,Wpercentile90,Ipercentile90,Opercentile90,
					Wpercentile50,Ipercentile50,Opercentile50,Wpercentile85,Ipercentile85,Opercentile85,
					stdCost, AvgCost, Cost99, Cost95, Cost85, Cost75,Cost50,Cost90) = scenario_test(s,q,xslp[0],N,n0,workbook2,recordsheet)
				worksheet.write(row,col+1,'SLP')
				worksheet.write(row,col+2,AvgWT)
				worksheet.write(row,col+3,AvgOT)
				worksheet.write(row,col+4,AvgIT)
				worksheet.write(row,col+5,AvgCost)
				row += 4
				worksheet.write(row,col+1,'SLP')
				worksheet.write(row,col+2,stdWT)
				worksheet.write(row,col+3,stdOT)
				worksheet.write(row,col+4,stdIT)
				worksheet.write(row,col+5,stdCost)
				row += 4
				worksheet.write(row,col+1,'SLP')
				worksheet.write(row,col+2,Wpercentile50)
				worksheet.write(row,col+3,Opercentile50)
				worksheet.write(row,col+4,Ipercentile50)				
				worksheet.write(row,col+5,Cost50)				
				row += 4
				worksheet.write(row,col+1,'SLP')
				worksheet.write(row,col+2,Wpercentile75)
				worksheet.write(row,col+3,Opercentile75)
				worksheet.write(row,col+4,Ipercentile75)
				worksheet.write(row,col+5,Cost75)				
				row += 4
				worksheet.write(row,col+1,'SLP')
				worksheet.write(row,col+2,Wpercentile85)
				worksheet.write(row,col+3,Opercentile85)
				worksheet.write(row,col+4,Ipercentile85)
				worksheet.write(row,col+5,Cost85)
				row += 4
				worksheet.write(row,col+1,'SLP')
				worksheet.write(row,col+2,Wpercentile90)
				worksheet.write(row,col+3,Opercentile90)
				worksheet.write(row,col+4,Ipercentile90)
				worksheet.write(row,col+5,Cost90)
							
				row += 4
				worksheet.write(row,col+1,'SLP')
				worksheet.write(row,col+2,Wpercentile95)
				worksheet.write(row,col+3,Opercentile95)
				worksheet.write(row,col+4,Ipercentile95)
				worksheet.write(row,col+5,Cost95)				
				row += 4
				worksheet.write(row,col+1,'SLP')
				worksheet.write(row,col+2,Wpercentile99)
				worksheet.write(row,col+3,Opercentile99)
				worksheet.write(row,col+4,Ipercentile99)
				worksheet.write(row,col+5,Cost99)				
	workbook2.close()			
	workbook.close()			
	
if __name__ == '__main__':
	main()