############################## COPYRIGHT ##############################
# All rights reserved. The authorized use of this data/code is limited to informational and educational purposes only, and NOT for operational or commercial purposes. No part of this zip file 

# including all the data generation files, example data output and code files may be reproduced, distributed, or transmitted in any form or by any means, without the prior written permission of Prof. 

# Siqian Shen (siqian@umich.edu). For permission requests, write to Siqian Shen (siqian@umich.edu), with a subject “Attention: Permissions for Using DR-noshow Data.”
########################################################################

# test_CP.py -- solve by using cutting-plane method with/without valid inequalities
# 	INPUT: 
# 	-- R-list of R values 
# 	-- p0-list of expected number of showing-up
# 	-- n0_list-list of number of appointments
# 	-- K-list of numbers of consecutive appts ruled out
# 	-- train_name-file name of training data

# 	OUTPUT:
# 	-- CPUtime_CP_n%s.xlsx-a spread sheet for each n0
# 	-- CPn%sK%sp%sR%s_%s.mat -- a .mat file containing solutions for each n0, K,R

from gurobipy import *
import math
import numpy
from numpy import *
from numpy.linalg import *
from numpy.linalg import inv
import csv
import scipy.io as sio
from scipy.stats import norm
from itertools import chain
import xlsxwriter

timelimit = 3600.0*1.5
timelimit2 = 3600.0*1.5	# time limit for solving without valid inequalities

# input parameters
R = [0]	# list of R values 
p0 = [0.6]	# list of expected number of showing-up
n0_list = [10,20,30,40,50]	# list of number of appointments



# function to solve without valid inequalities
# input parameters: su0, sl0, mu, sigma, n, nv, cu, cw, co, T, K (following notations in the paper), objCP -- objective value from solving with valid inequalities, for comparison purpose
# output: x_sol[0:n] -- solutions,t1 -- solving time,m1.objVal -- objective value,m1.status -- final master problem status,iteration -- number of iterations,t_MP -- time of solving master problems,T_SB -- time of solving subproblems,m2.NumConstrs -- number of constraints in the subproblem,node,opt_gap -- optimality gap
def CP_noVC(su0,sl0,mu,sigma,n,nv,cu,cw,co,T,K,objCP):
	nv0 = nv

	nv1 = floor(nv0*float(n))/float(n)
	nv2 = ceil(nv0*float(n))/float(n)

	pi = zeros((n+1,n+1))
	for i in range(n):
		for j in range(i,n):
			pi[i][j] = -cu[j]+sum(cw[i+1:j+1])
		pi[i][n]=sum(cw[i+1:n])+co
	sl = zeros((n+1,1))
	su = zeros((n+1,1))
	sl[0:n] = sl0
	su[0:n] = su0
	
	node = []	# record number of node branched
	t_MP = []	# record time of solving master problems
	T_SB = []	# record time of solving subproblems
	try:
  ##################### master problem ###########################3
		m1 = Model("lp")
		m1.setParam('OutputFlag',0)
		m1.setParam('DualReductions',0)
		m1.setParam('InfUnbdInfo',1)
		m1.params.timeLimit = timelimit2

		# define variables
		rho = []
		delta = m1.addVar(vtype=GRB.CONTINUOUS,lb=-GRB.INFINITY,obj=1.0,name='delta')
		x = []
		gamma = []
		
		for i in range(n+1):
	
			x.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0, name = 'x_%s'%(i)))
		for i in range(n):
			rho.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=-GRB.INFINITY, obj=mu[i], name='rho_%s'%(i)))
			gamma.append(m1.addVar(vtype = GRB.CONTINUOUS,lb=-GRB.INFINITY,obj = nv0[i],name = 'gamma_%s'%(i)))

		al = []
		ze = []
	
		phi = []
		tauu = []
		taul = []
		sigm = []
		lamb = []
		psil = []
		psiu = []
	
		for i in range(n+1):
			al.append(m1.addVar(vtype=GRB.CONTINUOUS, lb=-GRB.INFINITY, name='al_%s'%(i)))
		for i in range(n+1):
			tauu.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0, name = 'tauu_%s'%(i)))
		for i in range(n+1):
			taul.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0, name='taul_%s'%(i)))
		for i in range(n):
			ze.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=-GRB.INFINITY, name = 'ze_%s'%(i)))
		for i in range(n+1):			
			
			sigm.append([])
			lamb.append([])
			psil.append([])
			psiu.append([])
			for j in range(n+1):
				sigm[i].append([])
				lamb[i].append([])
				psiu[i].append([])
				psil[i].append([])
				for k in range(n+1):
					if (j>=k or i<j or i>=k):
						lamb[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'lamb_%s_%s_%s'%(i,j,k)))
					elif (i==n-1 and k==n and 0<=j<=n-1):
						lamb[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'lamb_%s_%s_%s'%(i,j,k)))
					else:
						lamb[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'lamb_%s_%s_%s'%(i,j,k)))
					if (j>k or i<j or i>k):
						sigm[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'sigm_%s_%s_%s'%(i,j,k)))
						psiu[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'pisu_%s_%s_%s'%(i,j,k)))
						psil[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'pisl_%s_%s_%s'%(i,j,k)))
					else:
						sigm[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'sigm_%s_%s_%s'%(i,j,k)))
						psiu[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'pisu_%s_%s_%s'%(i,j,k)))
						psil[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb = 0.0,name = 'pisl_%s_%s_%s'%(i,j,k)))

		
		for i in range(n):
			phi.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'phi_%s'%(i)))
		
		
		# Integrate new variables
		m1.update()
		m1.modelSense = GRB.MINIMIZE


		# integrate constraints

		m1.addConstr(quicksum(x[i] for i in range(n))==T[0])
		m1.addConstr(x[n]==0.0)
		
		
		for i in range(n):

			m1.addConstr(delta + quicksum(sl0[l][0]*rho[l] for l in range(n))+(su[i][0]-sl[i][0])*rho[i]+quicksum(nv1[l][0]*gamma[l] for l in range(n))>=0.0)
			m1.addConstr(delta + quicksum(sl0[l][0]*rho[l] for l in range(n))+quicksum(nv1[l][0]*gamma[l] for l in range(n)) + (nv2[i][0]-nv1[i][0])*gamma[i]>=0.0)
		
		m1.addConstr(quicksum(al[i]+su[i][0]*tauu[i]-sl[i][0]*taul[i] for i in range(n+1)) <= delta)
		m1.addConstr(phi[n-1]==0.0)
	
		
		for k in range(n+1):
			for j in range(k,n+1):
		
				tmp2 = quicksum(lamb[l][k][j] for l in range(k,j))
				tmp3 = sum(phi[j:min(j+1,n)])
				m1.addConstr(quicksum(al[l] for l in range(k,j+1)) - quicksum(sigm[l][k][j] for l in range(k,j+1)) + tmp2+ tmp3>= - quicksum(pi[l][j]*x[l] for l in range(k,j+1)))
		
		for i in range(n):
			m1.addConstr(ze[i]<= gamma[i])
			m1.addConstr(taul[i]-tauu[i]<=rho[i])
		
		for k in range(n+1):
			for i in range(k,n):
				for j in range(i,n+1):
					iu1 = min(j-1,i) + 1
					iu2 = min(n-1,max(2*i-j, i-1)) + 1
					il1 = max(k,i-1)
					il2 = max(0,min(2*i-k-1,i))
					tmp1 = quicksum(lamb[l][k][j] for l in range(il1,iu1))
					tmp2 = quicksum(phi[l] for l in range(il2,iu2))
					m1.addConstr(sigm[i][k][j]+sl[i][0]*psil[i][k][j]-su[i][0]*psiu[i][k][j]+ze[i]-sl[i][0]*taul[i]+su[i][0]*tauu[i]-tmp1-tmp2>=0.0)
	
		for k in range(n+1):
			for i in range(k,n):
				for j in range(i,n+1):	
					m1.addConstr(-psil[i][k][j] + psiu[i][k][j] + taul[i] - tauu[i]>= pi[i][j])

		m1.optimize()

		if m1.status == GRB.status.OPTIMAL:

			# read optimal solutions

			t1 = m1.runtime
			t_MP.append(t1)
			x_sol = []
			rho_sol = []
			gamma_sol = []
			delta_sol = m1.getVarByName('delta').x
			for i in range(0,n):
				rho_sol.append(m1.getVarByName('rho_%s'%(i)).x)
				gamma_sol.append(m1.getVarByName('gamma_%s'%i).x)
			for i in range(0,n+1):
				x_sol.append(m1.getVarByName('x_%s'%(i)).x)
			
		else:
			print "Initial master problem:"
			print m1.status

		

  ############subproblem##################################

		m2 = Model("mip")
		m2.setParam('OutputFlag',0)

		o = []
		t = []
		var_p = []
		var_q = []
		s = []
		for i  in range(n+1):
			o.append([])
			t.append([])
			var_p.append([])
			var_q.append(m2.addVar(vtype=GRB.BINARY,name='q_%s'%i))
			s.append(m2.addVar(vtype=GRB.CONTINUOUS,lb=sl[i][0],ub=su[i][0],name='s_%s'%i))
			for j in range(n+1):
				o[i].append([])
				t[i].append(m2.addVar(vtype=GRB.BINARY, name='t_%s_%s'%(i,j)))
				var_p[i].append([])
				for k in range(n+1):
					o[i][j].append(m2.addVar(vtype = GRB.CONTINUOUS,lb=0.0,name='o_%s_%s_%s'%(i,j,k)))
					var_p[i][j].append(m2.addVar(vtype=GRB.CONTINUOUS, lb = 0.0,name='p_%s_%s_%s'%(i,j,k)))
			
	# Integrate new variables
		m2.update()
		m2.modelSense = GRB.MAXIMIZE
		m2.addConstr(var_q[n]==0.0)
		m2.addConstr(s[n]==0.0)
		for i in range(n+1):
			m2.addConstr(quicksum(quicksum(t[k][j] for j in range(i,n+1)) for k in range(i+1))==1.0)
		for k in range(n+1):
			for j in range(k,n+1):
				for i in range(k,j+1):
					m2.addConstr(var_p[i][k][j]-t[k][j]<=0.0)
					m2.addConstr(var_p[i][k][j]-var_q[i]<=0.0)
					m2.addConstr(var_p[i][k][j] - var_q[i]-t[k][j]>=-1.0)

					m2.addConstr(o[i][k][j]-sl[i][0]*var_p[i][k][j]>=0.0)
					m2.addConstr(o[i][k][j]-su[i][0]*var_p[i][k][j]<=0.0)
					m2.addConstr(o[i][k][j]-s[i]+sl[i][0]*(1-var_p[i][k][j])<=0.0)
					m2.addConstr(o[i][k][j]-s[i]+su[i][0]*(1-var_p[i][k][j])>=0.0)

		for i in range(n-K+1):
			m2.addConstr(quicksum(var_q[j] for j in range(i,i+K-1+1))>=1.0)
  		
  # #######model end##############################################
	


		
		X_list = []	# record master problems' solutions
		X_list.append(x)

		sub_list = []
	
		find = 0	# =1 if find optimal, =0 if not
		iteration = 0 	# record iteration number
		t2 = 0.0	# subproblem runtime

		while find==0:

			if t1>timelimit:
				break

			iteration += 1
			
			obj = -(quicksum(rho_sol[l]*s[l] for l in range(n))+quicksum(gamma_sol[l]*var_q[l] for l in range(n)))
			for k in range(n+1):
				for j in range(k,n+1):
					for i in range(k,j+1):
						obj += pi[i][j]*o[i][k][j]-x_sol[i]*pi[i][j]*t[k][j]
			m2.setObjective(obj)
			m2.params.timeLimit = timelimit - t1
			m2.update()
			m2.optimize()
			t1 += m2.runtime

			t2 = m2.runtime
			T_SB.append(m2.runtime)
			node.append(m2.getAttr('NodeCount'))				
			sub_list.append(m2.objVal)
				
			if m2.status == GRB.status.OPTIMAL:

				t_sol = []
				o_sol = []
				s_sol = []
				var_q_sol = []
				for i in range(n+1):
					t_sol.append([])
					o_sol.append([])
					s_sol.append(m2.getVarByName('s_%s'%i).x)
					var_q_sol.append(m2.getVarByName('q_%s'%i).x)
					for j in range(n+1):
						t_sol[i].append(m2.getVarByName('t_%s_%s'%(i,j)).x)
						o_sol[i].append([])
						for  k in range(n+1):
							o_sol[i][j].append(m2.getVarByName('o_%s_%s_%s'%(i,j,k)).x)

				if delta_sol >= m2.objVal-1e-4:
					find = 1

				else:	# solve master problems with a cut generated from subproblem solutions
					obj = -(quicksum(rho[l]*s_sol[l] for l in range(n))+quicksum(gamma[l]*var_q_sol[l] for l in range(n)))
					
					for k in range(n+1):
						for j in range(k,n+1):
							for i in range(k,j+1):
								obj += pi[i][j]*o_sol[i][k][j]-x[i]*pi[i][j]*t_sol[k][j]
					m1.addConstr(obj<= delta)
					m1.optimize()
					t1 += m1.runtime

					if m1.status == GRB.status.OPTIMAL:						
						t_MP.append(m1.runtime)
						x_sol = []
						rho_sol = []
						gamma_sol = []
						delta_sol = m1.getVarByName('delta').x
						for i in range(0,n):
							rho_sol.append(m1.getVarByName('rho_%s'%(i)).x)
							gamma_sol.append(m1.getVarByName('gamma_%s'%i).x)
						for i in range(0,n+1):
							x_sol.append(m1.getVarByName('x_%s'%(i)).x)
						X_list.append(x_sol)
					else:
						print 'master problem: '
						print m1.status
		
			else:
				print 'subproblem:'
				print m2.status

		if t1 < timelimit:	
			opt_gap = -1
			return (x_sol[0:n],t1,m1.objVal,m1.status,iteration,t_MP,T_SB,m2.NumConstrs,node,opt_gap)
		else:
			opt_gap = objCP - m1.objVal
			return ([],t1,m1.objVal,999,iteration,t_MP,T_SB,m2.NumConstrs,node,opt_gap)
	
	except GurobiError as e:
		print ('Error reported: %s'% e.message)


# function to solve with valid inequalities
# input parameters: su0, sl0, mu, sigma, n, nv, cu, cw, co, T, K (following notations in the paper), objCP -- objective value from solving with valid inequalities, for comparison purpose
# output: x_sol[0:n] -- solutions,t1 -- solving time,m1.objVal -- objective value,m1.status -- final master problem status,iteration -- number of iterations,t_MP -- time of solving master problems,T_SB -- time of solving subproblems,m2.NumConstrs -- number of constraints in the subproblem,node,opt_gap -- optimality gap
def CP(su0,sl0,mu,sigma,n,nv,cu,cw,co,T,K):
	nv0 = nv

	nv1 = floor(nv0*float(n))/float(n)
	nv2 = ceil(nv0*float(n))/float(n)

	pi = zeros((n+1,n+1))
	for i in range(n):
		for j in range(i,n):
			pi[i][j] = -cu[j]+sum(cw[i+1:j+1])
		pi[i][n]=sum(cw[i+1:n])+co
	sl = zeros((n+1,1))
	su = zeros((n+1,1))
	sl[0:n] = sl0
	su[0:n] = su0
	
	node = []	# record number of node branched
	t_MP = []	# record time of solving master problems
	T_SB = []	# record time of solving subproblems

	try:

  #####################master problem###########################3
		m1 = Model("lp")
		m1.setParam('OutputFlag',0)
		m1.setParam('DualReductions',0)
		m1.setParam('InfUnbdInfo',1)
		m1.params.timeLimit = timelimit
	
		rho = []
		delta = m1.addVar(vtype=GRB.CONTINUOUS,lb=-GRB.INFINITY,obj=1.0,name='delta')
		x = []
		gamma = []
		
		for i in range(n+1):
	
			x.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0, name = 'x_%s'%(i)))
		for i in range(n):
			rho.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=-GRB.INFINITY, obj=mu[i], name='rho_%s'%(i)))
			gamma.append(m1.addVar(vtype = GRB.CONTINUOUS,lb=-GRB.INFINITY,obj = nv0[i],name = 'gamma_%s'%(i)))

		al = []
		ze = []
	
		phi = []
		tauu = []
		taul = []
		sigm = []
		lamb = []
		psil = []
		psiu = []
	
		for i in range(n+1):
			al.append(m1.addVar(vtype=GRB.CONTINUOUS, lb=-GRB.INFINITY, name='al_%s'%(i)))
		for i in range(n+1):
			tauu.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0, name = 'tauu_%s'%(i)))
		for i in range(n+1):
			taul.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0, name='taul_%s'%(i)))
		for i in range(n):
			ze.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=-GRB.INFINITY, name = 'ze_%s'%(i)))
		for i in range(n+1):			
			
			sigm.append([])
			lamb.append([])
			psil.append([])
			psiu.append([])
			for j in range(n+1):
				sigm[i].append([])
				lamb[i].append([])
				psiu[i].append([])
				psil[i].append([])
				for k in range(n+1):
					if (j>=k or i<j or i>=k):
						lamb[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'lamb_%s_%s_%s'%(i,j,k)))
					elif (i==n-1 and k==n and 0<=j<=n-1):
						lamb[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'lamb_%s_%s_%s'%(i,j,k)))
					else:
						lamb[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'lamb_%s_%s_%s'%(i,j,k)))
					if (j>k or i<j or i>k):
						sigm[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'sigm_%s_%s_%s'%(i,j,k)))
						psiu[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'pisu_%s_%s_%s'%(i,j,k)))
						psil[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'pisl_%s_%s_%s'%(i,j,k)))
					else:
						sigm[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'sigm_%s_%s_%s'%(i,j,k)))
						psiu[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'pisu_%s_%s_%s'%(i,j,k)))
						psil[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb = 0.0,name = 'pisl_%s_%s_%s'%(i,j,k)))

		
		for i in range(n):
			phi.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'phi_%s'%(i)))
		
		
		# Integrate new variables
		m1.update()
		m1.modelSense = GRB.MINIMIZE

		m1.addConstr(quicksum(x[i] for i in range(n))==T[0])
		m1.addConstr(x[n]==0.0)
		
		
		for i in range(n):

			m1.addConstr(delta + quicksum(sl0[l][0]*rho[l] for l in range(n))+(su[i][0]-sl[i][0])*rho[i]+quicksum(nv1[l][0]*gamma[l] for l in range(n))>=0.0)
			m1.addConstr(delta + quicksum(sl0[l][0]*rho[l] for l in range(n))+quicksum(nv1[l][0]*gamma[l] for l in range(n)) + (nv2[i][0]-nv1[i][0])*gamma[i]>=0.0)
	
		m1.addConstr(quicksum(al[i]+su[i][0]*tauu[i]-sl[i][0]*taul[i] for i in range(n+1)) <= delta)
		m1.addConstr(phi[n-1]==0.0)
	
		
		for k in range(n+1):
			for j in range(k,n+1):
		
				tmp2 = quicksum(lamb[l][k][j] for l in range(k,j))
				tmp3 = sum(phi[j:min(j+1,n)])
				m1.addConstr(quicksum(al[l] for l in range(k,j+1)) - quicksum(sigm[l][k][j] for l in range(k,j+1)) + tmp2+ tmp3>= - quicksum(pi[l][j]*x[l] for l in range(k,j+1)))
		
		for i in range(n):
			m1.addConstr(ze[i]<= gamma[i])
			m1.addConstr(taul[i]-tauu[i]<=rho[i])
		
		for k in range(n+1):
			for i in range(k,n):
				for j in range(i,n+1):
					iu1 = min(j-1,i) + 1
					iu2 = min(n-1,max(2*i-j, i-1)) + 1
					il1 = max(k,i-1)
					il2 = max(0,min(2*i-k-1,i))
					tmp1 = quicksum(lamb[l][k][j] for l in range(il1,iu1))
					tmp2 = quicksum(phi[l] for l in range(il2,iu2))
					m1.addConstr(sigm[i][k][j]+sl[i][0]*psil[i][k][j]-su[i][0]*psiu[i][k][j]+ze[i]-sl[i][0]*taul[i]+su[i][0]*tauu[i]-tmp1-tmp2>=0.0)
	
		for k in range(n+1):
			for i in range(k,n):
				for j in range(i,n+1):	
					m1.addConstr(-psil[i][k][j] + psiu[i][k][j] + taul[i] - tauu[i]>= pi[i][j])

		m1.optimize()

		if m1.status == GRB.status.OPTIMAL:
			t1 = m1.runtime
			t_MP.append(t1)
			x_sol = []
			rho_sol = []
			gamma_sol = []
			delta_sol = m1.getVarByName('delta').x
			for i in range(0,n):
				rho_sol.append(m1.getVarByName('rho_%s'%(i)).x)
				gamma_sol.append(m1.getVarByName('gamma_%s'%i).x)
			for i in range(0,n+1):
				x_sol.append(m1.getVarByName('x_%s'%(i)).x)
			
		else:
			print "initial master problem:"
			print m1.status
			
		

  ############subproblem##################################
		m2 = Model("mip")
		m2.setParam('OutputFlag',0)
		m2.params.timeLimit = timelimit

		o = []
		t = []
		var_p = []
		var_q = []
		s = []
		for i  in range(n+1):
			o.append([])
			t.append([])
			var_p.append([])
			var_q.append(m2.addVar(vtype=GRB.BINARY,name='q_%s'%i))
			s.append(m2.addVar(vtype=GRB.CONTINUOUS,lb=sl[i][0],ub=su[i][0],name='s_%s'%i))
			for j in range(n+1):
				o[i].append([])
				t[i].append(m2.addVar(vtype=GRB.BINARY, name='t_%s_%s'%(i,j)))
				var_p[i].append([])
				for k in range(n+1):
					o[i][j].append(m2.addVar(vtype = GRB.CONTINUOUS,lb=0.0,name='o_%s_%s_%s'%(i,j,k)))
					var_p[i][j].append(m2.addVar(vtype=GRB.CONTINUOUS, lb = 0.0,name='p_%s_%s_%s'%(i,j,k)))
			
	# Integrate new variables
		m2.update()
		m2.modelSense = GRB.MAXIMIZE
		m2.addConstr(var_q[n]==0)
		m2.addConstr(s[n]==0)
		for i in range(n+1):
			m2.addConstr(quicksum(quicksum(t[k][j] for j in range(i,n+1)) for k in range(i+1))==1.0)
		for k in range(n+1):
			for j in range(k,n+1):
				for i in range(k,j+1):
					m2.addConstr(var_p[i][k][j]-t[k][j]<=0.0)
					m2.addConstr(var_p[i][k][j]-var_q[i]<=0.0)
					m2.addConstr(var_p[i][k][j] - var_q[i]-t[k][j]>=-1.0)

					m2.addConstr(o[i][k][j]-sl[i][0]*var_p[i][k][j]>=0.0)
					m2.addConstr(o[i][k][j]-su[i][0]*var_p[i][k][j]<=0.0)
					m2.addConstr(o[i][k][j]-s[i]+sl[i][0]*(1-var_p[i][k][j])<=0.0)
					m2.addConstr(o[i][k][j]-s[i]+su[i][0]*(1-var_p[i][k][j])>=0.0)

		for i in range(n-K+1):
			m2.addConstr(quicksum(var_q[j] for j in range(i,i+K-1+1))>=1.0)
  		
  		for i in range(n+1):
  			m2.addConstr(quicksum(quicksum(var_p[i][k][j] for j in range(i,n+1)) for k in range(i+1))==var_q[i])
  			m2.addConstr(s[i]-quicksum(quicksum(o[i][k][j] - sl[i][0]*var_p[i][k][j] for j in range(i,n+1)) for k in range(i+1))>=sl[i][0])
  			m2.addConstr(s[i]-quicksum(quicksum(o[i][k][j] - su[i][0]*var_p[i][k][j] for j in range(i,n+1)) for k in range(i+1))<=su[i][0])
  		
  		for k in range(n+1):
  			for j in range(k+1,n+1):
  				for i in range(k,j-K+2):
  					m2.addConstr(quicksum(var_p[l][k][j] for l in range(i,i+K))>= t[k][j])
  		for i in range(K-1-1,n):
  			m2.addConstr(quicksum(quicksum(var_p[l][k][i] for l in range(i-K+2,i+1)) for k in range(i-K+2+1))+quicksum(var_p[i+1][i+1][j] for j in range(i-K+2+1,n+1))>=quicksum(t[k][i] for k in range(i-K+2+1)))
  		for i in range(n-K+2):
  			m2.addConstr(quicksum(var_p[i][k][i] for k in range(i+1))+quicksum(quicksum(var_p[l][i+1][j] for j in range(i+K-1,n+1)) for l in range(i+1,i+K))>= quicksum(t[i+1][j] for j in range(i+K-1,n+1)))
  
  #######model end##############################################
	
	
		X_list = []	# record master problems' solutions
		X_list.append(x)	
		sub_list = []	# record objective values of subproblems 

		find = 0	# =1 find the optimal solution, =0 otherwise
		iteration = 0	# number of iterations

		while find==0:
			if t1>timelimit:
				break

			iteration += 1
			
			obj = -(quicksum(rho_sol[l]*s[l] for l in range(n))+quicksum(gamma_sol[l]*var_q[l] for l in range(n)))
			for k in range(n+1):
				for j in range(k,n+1):
					for i in range(k,j+1):
						obj += pi[i][j]*o[i][k][j]-x_sol[i]*pi[i][j]*t[k][j]
			m2.setObjective(obj)
			m2.update()
			m2.optimize()
			T_SB.append(m2.runtime)
			t1 += m2.runtime
			sub_list.append(m2.objVal)	

			if m2.status == GRB.status.OPTIMAL:
				node.append(m2.getAttr('NodeCount'))

				# m1.write('bug.lp')
				t_sol = []
				o_sol = []
				s_sol = []
				var_q_sol = []
				for i in range(n+1):
					t_sol.append([])
					o_sol.append([])
					s_sol.append(m2.getVarByName('s_%s'%i).x)
					var_q_sol.append(m2.getVarByName('q_%s'%i).x)
					for j in range(n+1):
						t_sol[i].append(m2.getVarByName('t_%s_%s'%(i,j)).x)
						o_sol[i].append([])
						for  k in range(n+1):
							o_sol[i][j].append(m2.getVarByName('o_%s_%s_%s'%(i,j,k)).x)

				if delta_sol >= m2.objVal-1e-4:
					find = 1
				else:
					obj = -(quicksum(rho[l]*s_sol[l] for l in range(n))+quicksum(gamma[l]*var_q_sol[l] for l in range(n)))
					
					for k in range(n+1):
						for j in range(k,n+1):
							for i in range(k,j+1):
								obj += pi[i][j]*o_sol[i][k][j]-x[i]*pi[i][j]*t_sol[k][j]
					m1.addConstr(obj<= delta)
					m1.optimize()
					
					if m1.status == GRB.status.OPTIMAL:
						t1 += m1.runtime
						t_MP.append(m1.runtime)
						x_sol = []
						rho_sol = []
						gamma_sol = []
						delta_sol = m1.getVarByName('delta').x
						for i in range(0,n):
							# x_sol.append(m1.getVarByName('x_%s'%(i)).x)
							rho_sol.append(m1.getVarByName('rho_%s'%(i)).x)
							gamma_sol.append(m1.getVarByName('gamma_%s'%i).x)
						for i in range(0,n+1):
							x_sol.append(m1.getVarByName('x_%s'%(i)).x)
						X_list.append(x_sol)
					else:
						print 'infeasible'
		
			else:
				print 'subproblem: infeasible'
		if t1 < timelimit:	
			
			return (x_sol[0:n],t1,m1.objVal,m1.status,iteration,t_MP,T_SB,m2.NumConstrs,node)
		else:
			return ([],t1,0,999,iteration,t_MP,T_SB,m2.NumConstrs,node)
	
		
	except GurobiError as e:
		print ('Error reported: %s'% e.message)


# function to read data from train_file
def readfile(train_file):
	mat = sio.loadmat(train_file)
	su = mat['su'].astype(float).tolist()
	sl = mat['sl'].astype(float).tolist()
	
	mu = mat['mu'].astype(float)
	s = mat['s'].astype(float)
	sigma = mat['sigma'].astype(float)
	n = mat['n']
	N = mat['N']
	q = mat['q'].astype(float)
	nv = mat['nv'].astype(float)
	return (su,sl,s,mu,sigma,n,N,q,nv)


def main():

	
	
	for  ni in range(len(n0_list)):
		n0 = n0_list[ni]
		workbook = xlsxwriter.Workbook('CPUtime_CP_n%s.xlsx'%n0)
		worksheet = workbook.add_worksheet()
		title = ['','','','Obj', ' ','CPU','','# iteration','','avg. time of sub ','','avg. time of mp','','# constraints','','status','','avg. # of node','','opt_gap']
		title0 = ['K','R','nv','CP_noVC','CP','CP_noVC','CP','CP_noVC','CP','CP_noVC','CP','CP_noVC','CP','CP_noVC','CP','CP_noVC','CP','CP_noVC','CP','CP_noVC']
		title1 = ['CP_noVC','iter','Sub','MP',' # of Node','CP', 'iter','Sub','MP', '# of Node']
		for i in range(len(title)):
			worksheet.write(0,i,title[i])
			worksheet.write(1,i,title0[i])	
		
# INPUT
		# number of consecutive appts ruled out
		# can be defined independently
		K = [int(floor(.3*n0)),int(floor(.7*n0))]
		
		cu = 0.5*ones((n0,1))
		cw = ones((n0,1))
		co = 10.0
	
		row = 2

		for k in range(len(K)):
			col = 0
			worksheet.write(row,col,K[k])
			col += 1
			for tt in range(1,2):	
				worksheet1 = workbook.add_worksheet('K=%s_%s'%(K[k],tt))
				for i  in range(len(title1)):
					worksheet1.write(0,i,title1[i])
					
				for tR in  range(len(R)):
					
					for ts in range(len(p0)):
				
						worksheet.write(row,col,R[tR])
						worksheet.write(row,col+1,p0[ts])
						col += 2

# INPUT
						train_name = './train_data/%sseed_1465265434_%sR.mat'%(int(n0),4)
						(su,sl,s,mu,sigma,n,N,q,nv) = readfile(train_name)
						n = n[0][0]

						
						T = sum(mu)+R[tR]*sqrt(sum(sigma[i]*sigma[i] for i in range(n0)))
						
						(x,time,objVal,status,iteration,t_MP,T_SB,nc,node) = CP(su,sl,mu,sigma,n0,nv,cu,cw,co,T,int(K[k]))
						print 'Cutting Plane done'
						objCP = objVal
						
						(x_noVC,time_noVC,obj_noVC,status_noVC,iteration_noVC,t_MP_noVC,T_SB_noVC,nc_noVC,node_noVC,opt_gap) = CP_noVC(su,sl,mu,sigma,n0,nv,cu,cw,co,T,int(K[k]),objCP)
						
						print 'Cutting Plane without Valid Inequalities done'	
						print objVal,obj_noVC,opt_gap		
						worksheet.write(row,col,obj_noVC)
						col +=2
						
						worksheet.write(row,col,time_noVC)
						worksheet.write(row,col+2,iteration_noVC)
						worksheet.write(row,col+4,mean(T_SB_noVC))
						worksheet.write(row,col+6,mean(t_MP_noVC))
						worksheet.write(row,col+8,nc_noVC)
						worksheet.write(row,col+10,status_noVC)
						worksheet.write(row,col+12,mean(node_noVC))
						worksheet.write(row,col+14,opt_gap)
						col -= 1
				
						for  i in range(iteration_noVC):
							worksheet1.write(i+1,1,i+1)
							worksheet1.write(i+1,2,T_SB_noVC[i])
							worksheet1.write(i+1,3,t_MP_noVC[i])
							worksheet1.write(i+1,4,node_noVC[i])
						
						
						worksheet.write(row,col,objVal)
						col += 2
						worksheet.write(row,col,time)
						worksheet.write(row,col+2,iteration)
						worksheet.write(row,col+4,mean(T_SB))
						worksheet.write(row,col+6,mean(t_MP))
						worksheet.write(row,col+8,nc)
						worksheet.write(row,col+10,status)
						worksheet.write(row,col+12,mean(node))
						
						col -= 1
				
						for  i in range(iteration):
							worksheet1.write(i+1,1+5,i+1)
							worksheet1.write(i+1,2+5,T_SB[i])
							worksheet1.write(i+1,3+5,t_MP[i])
							worksheet1.write(i+1,4+5,node[i])

					sio.savemat('./solution/CPn%sK%sp%sR%s_%s.mat'%(n0,K[k],p0[ts],R[tR],tt),{'x':x,'xnoVC':x_noVC})
			
					row += 1


		workbook.close()
					
		


			
if __name__ == '__main__':
	main()