############################## COPYRIGHT ##############################
# All rights reserved. The authorized use of this data/code is limited to informational and educational purposes only, and NOT for operational or commercial purposes. No part of this zip file 

# including all the data generation files, example data output and code files may be reproduced, distributed, or transmitted in any form or by any means, without the prior written permission of Prof. 

# Siqian Shen (siqian@umich.edu). For permission requests, write to Siqian Shen (siqian@umich.edu), with a subject “Attention: Permissions for Using DR-noshow Data.”
########################################################################

# test_ED.py -- solve by using ED(2) and ED(n+1) models
# 	INPUT: 
# 	-- R-list of R values 
# 	-- p0-list of expected number of showing-up
# 	-- n0_list-list of number of appointments
# 	-- train_name-file name of training data

# 	OUTPUT:
# 	-- CPUtime_CPED_n%s.xlsx-a spread sheet for each n0
# 	-- CPEDn%sp%sR%s_%s.mat -- a .mat file containing solutions for each n0,R

from gurobipy import *
import math
import numpy
from numpy import *
from numpy.linalg import *
from numpy.linalg import inv
import csv
import scipy.io as sio
from scipy.stats import norm
from itertools import chain
import xlsxwriter

timelimit = 3600.0*3	# time limit for solving ED2 or ED(n+1) model

R = [0]	# list of R values 
p0 = [0.6]	# list of expected number of showing-up
n0_list = [50]	# list of number of appointments
	

# function to solve ED(2) model
def  ED2(su0,sl0,mu,sigma,n,nv,cu,cw,co,T):
	nv0 = nv
	ep = 0.0#0.338	
	pi = zeros((n+1,n+1))
	for i in range(n):
		for j in range(i,n):
			pi[i][j] = -cu[j]+sum(cw[i+1:j+1])
		pi[i][n]=sum(cw[i+1:n])+co
	sl = zeros((n+1,1))
	su = zeros((n+1,1))

	sl[0:n] = sl0[0:n]
	su[0:n] = su0[0:n]
	try:

		m1 = Model("lp")
		
		m1.setParam('OutputFlag',0)
		m1.setParam('DualReductions',0)
		
		rho = []
		al = []
		ze = []
		x = []
		phi = []
		tauu = []
		taul = []
		sigm = []
		lamb = []
		psil = []
		psiu = []
		gamma = []
		for i in range(n+1):
			al.append(m1.addVar(vtype=GRB.CONTINUOUS,obj = 1.0, lb=-GRB.INFINITY, name='al_%s'%(i)))
		for i in range(n+1):
			tauu.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,obj = su[i][0], name = 'tauu_%s'%(i)))
		for i in range(n+1):
			taul.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,obj = -sl[i][0], name='taul_%s'%(i)))
		for i in range(n):
			ze.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=-GRB.INFINITY, name = 'ze_%s'%(i)))
		for i in range(n+1):			
			x.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0, name = 'x_%s'%(i)))
			
			sigm.append([])
			lamb.append([])
			psil.append([])
			psiu.append([])
			for j in range(n+1):
				sigm[i].append([])
				lamb[i].append([])
				psiu[i].append([])
				psil[i].append([])
				for k in range(n+1):
					if (j>=k or i<j or i>=k):
						lamb[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'lamb_%s_%s_%s'%(i,j,k)))
					elif (i==n-1 and k==n and 0<=j<=n-1):
						lamb[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'lamb_%s_%s_%s'%(i,j,k)))
					else:
						lamb[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'lamb_%s_%s_%s'%(i,j,k)))
					if (j>k or i<j or i>k):
						sigm[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'sigm_%s_%s_%s'%(i,j,k)))
						psiu[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'pisu_%s_%s_%s'%(i,j,k)))
						psil[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,ub=0.0,name = 'pisl_%s_%s_%s'%(i,j,k)))
					else:
						sigm[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'sigm_%s_%s_%s'%(i,j,k)))
						psiu[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'pisu_%s_%s_%s'%(i,j,k)))
						psil[i][j].append(m1.addVar(vtype=GRB.CONTINUOUS,lb = 0.0,name = 'pisl_%s_%s_%s'%(i,j,k)))

		for i in range(n):
			rho.append(m1.addVar(vtype=GRB.CONTINUOUS, lb=-GRB.INFINITY,obj=mu[i], name='rho_%s'%(i)))
			gamma.append(m1.addVar(vtype = GRB.CONTINUOUS,lb=-GRB.INFINITY,obj = nv0[i],name = 'gamma_%s'%(i)))
		
		for i in range(n):
			phi.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name = 'phi_%s'%(i)))
			

		m1.update()
		m1.modelSense = GRB.MINIMIZE

		m1.addConstr(quicksum(x[i] for i in range(n))==T[0])
		m1.addConstr(x[n]==0.0)	
		m1.addConstr(phi[n-1]==0.0)
	
		
		for k in range(n+1):
			for j in range(k,n+1):
		
				tmp2 = quicksum(lamb[l][k][j] for l in range(k,j))
				tmp3 = sum(phi[j:min(j+1,n)])
				m1.addConstr(quicksum(al[l] for l in range(k,j+1)) - quicksum(sigm[l][k][j] for l in range(k,j+1)) + tmp2+ tmp3>= - quicksum(pi[l][j]*x[l] for l in range(k,j+1))-ep)
		
		for i in range(n):
			m1.addConstr(ze[i]-ep<= gamma[i])
			m1.addConstr(taul[i]-tauu[i]-ep<=rho[i])
		
		for k in range(n+1):
			for i in range(k,n):
				for j in range(i,n+1):
					iu1 = min(j-1,i) + 1
					iu2 = min(n-1,max(2*i-j, i-1)) + 1
					il1 = max(k,i-1)
					il2 = max(0,min(2*i-k-1,i))
					tmp1 = quicksum(lamb[l][k][j] for l in range(il1,iu1))
					tmp2 = quicksum(phi[l] for l in range(il2,iu2))
					m1.addConstr(sigm[i][k][j]+sl[i][0]*psil[i][k][j]-su[i][0]*psiu[i][k][j]+ze[i]-sl[i][0]*taul[i]+su[i][0]*tauu[i]-tmp1-tmp2>=0.0-ep)
	
		for k in range(n+1):
			for i in range(k,n):
				for j in range(i,n+1):	
					m1.addConstr(-psil[i][k][j] + psiu[i][k][j] + taul[i] - tauu[i]>= pi[i][j]-ep)
	
		m1.optimize()
		runtime =  m1.runtime
		
		if m1.status == GRB.status.OPTIMAL:
	
			m1.write('bug.lp')
			x = []

			for i in range(0,n):
				x.append(m1.getVarByName('x_%s'%(i)).x)
			
			return (x,runtime,m1.objVal,m1.status)
		else:
			return([],0,0,m1.status)
	
	except GurobiError as e:
		print ('Error reported: %s'% e.message)

# function to solve ED(n+1) model
def  EDn(su0,sl0,mu,sigma,n,nv,cu,cw,co,T):
	
	nv0 = nv
	
	pi = zeros((n+1,n+1))
	for i in range(n):
		for j in range(i,n):
			pi[i][j] = -cu[j]+sum(cw[i+1:j+1])
		pi[i][n]=sum(cw[i+1:n])+co
	sl = zeros((n+1,1))
	su = zeros((n+1,1))
	sl[0:n] = sl0
	su[0:n] = su0
	try:

		m1 = Model("lp")
		m1.setParam('OutputFlag',0)
		rho = []
		al = []
		x = []
		gamma = []
		beta = []
		for i in range(n+1):
			al.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=-GRB.INFINITY,obj = 1.0, name='al_%s'%(i)))
			x.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0, name = 'x_%s'%(i)))
			beta.append([])			
			for j in range(n+1):
				beta[i].append(m1.addVar(vtype=GRB.CONTINUOUS,lb=-GRB.INFINITY,name = 'beta_%s_%s'%(i,j)))
		for i in range(n):
			rho.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=-GRB.INFINITY, obj=mu[i], name='rho_%s'%(i)))
			gamma.append(m1.addVar(vtype = GRB.CONTINUOUS,lb=-GRB.INFINITY,obj = nv0[i],name = 'gamma_%s'%(i)))
		

		# Integrate new variables
		m1.update()
		m1.modelSense = GRB.MINIMIZE

		m1.addConstr(quicksum(x[i] for i in range(n))==T[0])
		m1.addConstr(x[n]==0)
		
		for k in range(n+1):
			for j in range(k,n+1):
				m1.addConstr(quicksum(al[i] for i in range(k,j+1))>= quicksum(beta[i][j] for i in range(k,j+1)))
		m1.addConstr(beta[n][n]==0)	

		for i in range(n):
			for j in range(i,n+1):
				m1.addConstr(beta[i][j]>=-pi[i][j]*x[i]-sl[i][0]*rho[i])
				m1.addConstr(beta[i][j]>=-pi[i][j]*x[i]-su[i][0]*rho[i])
				m1.addConstr(beta[i][j]>=-pi[i][j]*x[i]-sl[i][0]*rho[i]-gamma[i]+sl[i][0]*pi[i][j])
				m1.addConstr(beta[i][j]>=-pi[i][j]*x[i]-su[i][0]*rho[i]-gamma[i]+su[i][0]*pi[i][j])
		m1.optimize()
		runtime = m1.runtime
		if m1.status == GRB.status.OPTIMAL:
			m1.write('EDn.lp')			
			x = []
			for i in range(0,n):
				x.append(m1.getVarByName('x_%s'%(i)).x)
				
			return (x,runtime,m1.objVal,m1.status)
		else:
			return([],0,0,m1.status)
	
	except GurobiError as e:
		print ('Error reported: %s'% e.message)

# function to read training data from train_file
def readfile(train_file):
	mat = sio.loadmat(train_file)
	su = mat['su'].astype(float).tolist()
	sl = mat['sl'].astype(float).tolist()
	
	mu = mat['mu'].astype(float)
	s = mat['s'].astype(float)
	sigma = mat['sigma'].astype(float)
	n = mat['n']
	N = mat['N']
	q = mat['q'].astype(float)
	nv = mat['nv'].astype(float)
	return (su,sl,s,mu,sigma,n,N,q,nv)


def main():
	
	
		
	for  ni in range(len(n0_list)):
		n0 = n0_list[ni]
		workbook = xlsxwriter.Workbook('CPUtime_CPED_n%s.xlsx'%n0)
		worksheet = workbook.add_worksheet()
		title = ['','','','Obj', ' ','CPU','','# iteration','','avg. time of sub ','','avg. time of mp','','# constraints','','status','','avg. # of node','','opt_gap']
		title0 = ['K','R','nv','ED2','EDn','ED2','EDn','ED2','EDn','ED2','EDn','ED2','EDn','ED2','EDn','ED2','EDn','ED2','EDn','ED2']
		title1 = ['CP_noVC','iter','Sub','MP',' # of Node','CP', 'iter','Sub','MP', '# of Node']
		for i in range(len(title)):
			worksheet.write(0,i,title[i])
			worksheet.write(1,i,title0[i])	
		
		cu = 0.5*ones((n0,1))
		cw = ones((n0,1))
		co = 10.0
	
		row = 2

		
		col = 0
		col += 1
		for tt in range(1,2):	
			worksheet1 = workbook.add_worksheet('%s'%(tt))
			for i  in range(len(title1)):
				worksheet1.write(0,i,title1[i])
				
			for tR in  range(len(R)):
				
				for ts in range(len(p0)):
			
					# col = 1
					worksheet.write(row,col,R[tR])
					worksheet.write(row,col+1,p0[ts])
					col += 2

# INPUT
					train_name = './train_data/%sseed_1465265434_%sR.mat'%(int(n0),4)
					(su,sl,s,mu,sigma,n,N,q,nv) = readfile(train_name)
					n = n[0][0]

					
					T = sum(mu)+R[tR]*sqrt(sum(sigma[i]*sigma[i] for i in range(n0)))
					
					(x,time,objVal,status) = EDn(su,sl,mu,sigma,n0,nv,cu,cw,co,T)
					# print time
					objCP = objVal
					(x_noVC,time_noVC,obj_noVC,status_noVC) = ED2(su,sl,mu,sigma,n0,nv,cu,cw,co,T)
					# print time_noVC
		
					
					worksheet.write(row,col,obj_noVC)
					col +=2
					worksheet.write(row,col,time_noVC)
					worksheet.write(row,col+10,status_noVC)
					col -= 1
			
					worksheet.write(row,col,objVal)
					col += 2
					worksheet.write(row,col,time)
					worksheet.write(row,col+10,status)
				
					
					col -= 1
			
				
				sio.savemat('./solution/CPEDn%sp%sR%s_%s.mat'%(n0,p0[ts],R[tR],tt),{'x':x,'xnoVC':x_noVC})
		
				row += 1


		workbook.close()
					
		

			
if __name__ == '__main__':
	main()