############################## COPYRIGHT ##############################
# All rights reserved. The authorized use of this data/code is limited to informational and educational purposes only, and NOT for operational or commercial purposes. No part of this zip file 

# including all the data generation files, example data output and code files may be reproduced, distributed, or transmitted in any form or by any means, without the prior written permission of Prof. 

# Siqian Shen (siqian@umich.edu). For permission requests, write to Siqian Shen (siqian@umich.edu), with a subject “Attention: Permissions for Using DR-noshow Data.”
########################################################################

# test_SLP.py -- solve by using SLP
# 	INPUT: 
# 	-- R-list of R values 
# 	-- p0-list of expected number of showing-up
# 	-- n0_list-list of number of appointments
# 	-- train_name-file name of training data

# 	OUTPUT:
# 	-- sol&obj_SLP.xlsx-a spread sheet containing CPU time and solution details
# 	-- SLPn%sp%sR%s_%s.mat -- a .mat file containing solutions for each n0, R, p


from gurobipy import *
import math
import numpy
from numpy import *
from numpy.linalg import *
from numpy.linalg import inv
import csv
import scipy.io as sio
from scipy.stats import norm
from itertools import chain
import xlsxwriter
import timeit


# function to solve by using SLP
def  SLP(s,mu,sigma,n,N,nv,cu,cw,co,T,q,t1):
	
	try:

		m1 = Model("lp")
		# m1.params.timeLimit = timelimit
		# m1.setParam('OutputFlag',0)
		# m1.setParam('Heuristics',0)
		# m1.setParam('Cuts', 0)
		# m1.setParam('Presolve', 0)
		# m1.setParam('Threads',1)
		x = []
		w = []
		u = []
		
		for i in range(n+1):
			x.append(m1.addVar(vtype=GRB.CONTINUOUS,lb=0.0,name='x_%s'%i))
		for i in range(N):
			u.append([])
			w.append([])
			for j in range(n):
				w[i].append(m1.addVar(vtype=GRB.CONTINUOUS,obj = cw[j]/N,lb=0.0,name='w_%s_%s'%(i,j)))
				u[i].append(m1.addVar(vtype=GRB.CONTINUOUS,obj= cu[j]/N,lb=0.0,name='u_%s_%s'%(i,j)))
			w[i].append(m1.addVar(vtype=GRB.CONTINUOUS,obj = co/N,lb=0.0,name='w_%s_%s'%(i,n)))


		# Integrate new variables
		m1.update()
		m1.modelSense = GRB.MINIMIZE

		m1.addConstr(quicksum(x[i] for i in range(n))==T[0])
		m1.addConstr(x[n]==0)

		for t in range(N):
			m1.addConstr(w[t][n]-u[t][n-1]==q[t][n-1]*s[t][n-1]+w[t][n-1]+quicksum(x[l] for l in range(n-1))-T)
			m1.addConstr(w[t][0]==0)
	
			for i in range(1,n):
				m1.addConstr(w[t][i]-u[t][i-1]==q[t][i-1]*s[t][i-1]+w[t][i-1]-x[i-1])
		
		t2 = timeit.default_timer() - t1	
		t3 = timeit.default_timer()
		print "processing time", t2
		m1.optimize()
		print "solving time", timeit.default_timer()-t3
		if m1.status == GRB.status.OPTIMAL:
			x = []
			for i in range(0,n):
				x.append(m1.getVarByName('x_%s'%(i)).x)
				
			return (x,m1.runtime,m1.objVal,m1.status)
		else:
			return([],0,0,m1.status)
	
	except GurobiError as e:
		print ('Error reported: %s'% e.message)





# function to calculate the in-sample statistics
def InSample_test(s,q,x,N,n,cw,cu,co):
	
	u = zeros((n,N))
	w = zeros((n,N))
	OT = zeros((1,N))
					
	for j in  range(1,n):
		for i in range(N):
			w[j][i] = max(0,q[i][j-1]*s[i][j-1]+w[j-1][i]-x[j-1])
			u[j-1][i] = max(0,x[j-1]-q[i][j-1]*s[i][j-1]-w[j-1][i])
	
	for i in range(N):
		u[n-1][i] = max(0,x[n-1] - q[i][n-1]*s[i][n-1]-w[n-1][i])
		OT[0][i] = u[n-1][i] - x[n-1] +q[i][n-1]*s[i][n-1] + w[n-1][i]

	TotalWT = sum(w,axis = 0)
	TotalIT = sum(u,axis = 0)
	TotalOT = sum(OT)

	AvgWT = mean(TotalWT)
	AvgIT = mean(TotalIT)
	AvgOT = TotalOT/N
	# quan = [75,95]
	return cu[0]*AvgIT+co*AvgOT+cw[0]*AvgWT
	# Wpercentile75 = percentile(TotalWT,quan[0])
	# Ipercentile75 = percentile(TotalIT,quan[0])
	# Opercentile75 = percentile(TotalOT,quan[0])

	# Wpercentile95 = percentile(TotalWT,quan[1])
	# Ipercentile95 = percentile(TotalIT,quan[1])
	# Opercentile95 = percentile(TotalOT,quan[1])

	# return (AvgWT,AvgIT,AvgOT,Wpercentile95,Wpercentile75,Ipercentile95,Ipercentile75,Opercentile95,Opercentile75)

def readfile(train_file):
	mat = sio.loadmat(train_file)
	su = mat['su'].astype(float).tolist()
	sl = mat['sl'].astype(float).tolist()
	
	mu = mat['mu'].astype(float)
	s = mat['s'].astype(float)
	sigma = mat['sigma'].astype(float)
	n = mat['n']
	N = mat['N']
	q = mat['q'].astype(float)
	nv = mat['nv'].astype(float)
	return (su,sl,s,mu,sigma,n,N,q,nv)

def main():

# INPUT
	R0 = [0.0]
	p0 = [0.8,0.6]
	# n0_list = [10,15,20,25,30,35,40,45,50]
	n0_list = [10,15]
	

	train_file  = './train_data/'
	
	title = ['n','R','nv','CPU: ED2','EDn','SLP','Obj: ED2','EDn','SLP','IS_Obj: ED2', 'EDn', 'SLP', 'status: ED2','EDn','SLP']
	title2 = ['R','','x1','x2','x3','x4','x5','x6','x7','x8','x9','x10']
	
	workbook = xlsxwriter.Workbook('sol&obj_SLP.xlsx')
	worksheet = workbook.add_worksheet()
	for i in range(len(title)):
		worksheet.write(0,i+0,title[i])
	for i in range(len(title)):
		worksheet.write(0,i+18,title[i])

	row = 0
	for  ni in range(len(n0_list)):
		n0 = n0_list[ni]
		row += 1	
		
		cu = 0.5*ones((n0,1))
		cw = 1.0*ones((n0,1))
		co = 10.0

			
		
		col0 = -18		
		col20 = -18	
		for tt in range(1,2):	
			for ts in range(len(p0)):
				
				row2 = 10
				col0 += 18
				col20 += 18

				
				
				row2 += 1

				for tR  in [0]:
					col = col0
					col2 = col20
					R = R0[tR]
			
				
				
					worksheet.write(row,col,n0)					
					col += 1

					
				
					worksheet.write(row,col,R)
					worksheet.write(row,col+1,p0[ts])
					col += 2

# INPUT
					train_name = train_file+str(n0)+'SLPseed_1465265434'+'_%sR.mat'%(10-int(10*p0[ts]))
					t1 = timeit.default_timer()
					(su,sl,s,mu,sigma,n,N,q,nv) = readfile(train_name)
					n = n[0][0]

					T = sum(mu)+R*sqrt(sum(sigma[i]*sigma[i] for i in range(n0)))
					col += 1

					col += 1
					

					
					(x_slp,time_slp,obj_slp,status_slp) = SLP(s,mu,sigma,n0,N,nv,cu,cw,co,T,q,t1)
					inobj_slp = InSample_test(s,q,x_slp,N,n,cw,cu,co)
					
					worksheet.write(row,col,time_slp)
					worksheet.write(row,col+3,obj_slp)
					worksheet.write(row,col+6,inobj_slp)
					worksheet.write(row,col+9,status_slp)
					col += 1
					
					sio.savemat('./solution/SLPn%sp%sR%s_%s.mat'%(n0,p0[ts],R,tt),{'xslp':x_slp,'T':T})
					
					
	workbook.close()



		
if __name__ == '__main__':
	main()